"""Module for initializing settings related to the built-in seneca logger
Functions:
-get_logger"""

import logging, coloredlogs
import os, sys
from os.path import dirname
from logging.handlers import RotatingFileHandler
import cilantro_ee

VALID_LVLS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
_LOG_LVL = os.getenv('LOG_LEVEL', None)
if _LOG_LVL:
    assert _LOG_LVL in VALID_LVLS, "Log level {} not in valid levels {}".format(_LOG_LVL, VALID_LVLS)
    _LOG_LVL = getattr(logging, _LOG_LVL)
else:
    _LOG_LVL = 1

logging.raiseExceptions = False
req_log = logging.getLogger('urllib3')
req_log.setLevel(logging.WARNING)
req_log.propagate = True


def get_main_log_path():
    import cilantro_ee

    root = cilantro_ee.__file__  # resolves to '/Users/davishaba/Developer/cilantro_ee/cilantro_ee/__init__.py'
    log_path = '/'.join(root.split('/')[:-2]) + '/logs/cilantro_ee.log'

    # Create log directory if it does not exist
    log_dir = os.path.dirname(log_path)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    return log_path

format = '%(asctime)s.%(msecs)03d %(name)s[%(process)d][%(processName)s] <{}> %(levelname)-2s %(message)s'.format(os.getenv('HOST_NAME', 'Node'))

"""
Custom Log Levels
"""


#     'INFO': 20,
#     'WARNING': 30,
#     'ERROR': 40,
#     'CRITICAL': 50,

CUSTOM_LVL = {
    'SPAM': 12,
    'TEST': 14,
    'DEBUG3': 15,
    'DEBUG2': 16,
    'DEBUG1': 17,
    'DEBUGV': 21,
    'NOTICE': 22,
    'FATAL': 99
    }

DEPRECATED_LEVELS = {
    'SOCKET': 23,
    'NOTICE': 24,
    'SUCCESS': 26,
    'SUCCESS2': 27,
    'IMPORTANT': 56,
    'IMPORTANT2': 57,
    'IMPORTANT3': 58
}

CUSTOM_LEVELS = {**CUSTOM_LVL, **DEPRECATED_LEVELS}


# CUSTOM_LEVELS = {
#     'SPAM': 12,
#     'DEBUGV': 13,
#     'SOCKET': 23,
#     'NOTICE': 24,
#     'SUCCESS': 26,
#     'SUCCESS2': 27,
#     'IMPORTANT': 56,
#     'IMPORTANT2': 57,
#     'IMPORTANT3': 58,
#     'FATAL': 9001,
#     'TEST': 9002,
#     }

for log_name, log_level in CUSTOM_LEVELS.items():
    logging.addLevelName(log_level, log_name)


def apply_custom_level(log, name: str, level: int):
    def _lvl_func(message, *args, **kws):
        if level >= log.getEffectiveLevel():
            log._log(level, message, args, **kws)

    setattr(log, name.lower(), _lvl_func)


"""
Custom Styling
"""

coloredlogs.DEFAULT_LEVEL_STYLES = {
    'critical': {'color': 'white', 'bold': True, 'background': 'red'},
    'fatal': {'color': 'white', 'bold': True, 'background': 'red', 'underline': True},
    'debug': {'color': 'green'},
    'error': {'color': 'red'},
    'info': {'color': 'white'},
    'test': {'color': 'magenta'},
    'warning': {'color': 'yellow'},
    'debugv': {'color': 'blue', 'faint': False},
    'debug1': {'color': 'cyan', 'bold': True, 'background': 'magenta'},
    'debug2': {'color': 'magenta', 'bold': True, 'background': 'cyan'},
    'debug3': {'color': 'black', 'bold': True, 'background': 'yellow'},

    # TODO FOLLOWING ARE DEPRECATED LEVELS WOULD BE REMOVED GOING FORWARD

    'notice': {'color': 'white', 'background': 'yellow'},
    'socket': {'color': 'white', 'background': 'yellow'},
    'important': {'color': 'white', 'background': 'yellow'},
    'important2': {'color': 'white', 'background': 'yellow'},
    'important3': {'color': 'white', 'background': 'yellow'},
    'spam': {'color': 'white', 'background': 'yellow'},
    'success': {'color': 'white', 'background': 'yellow'},
    'success2': {'color': 'white', 'background': 'yellow'},
    'verbose': {'color': 'white', 'background': 'yellow'}
}
coloredlogs.DEFAULT_FIELD_STYLES = {
    'asctime': {'color': 'green'},
    'hostname': {'color': 'magenta'},
    'levelname': {'color': 'black', 'bright': True},
    'name': {'color': 'blue'},
    'programname': {'color': 'cyan'}
}


class LoggerWriter:
    def __init__(self, level):
        self.level = level
    def write(self, message):
        if message != '\n':
            self.level(message)
    def flush(self):
        return


class ColoredFileHandler(RotatingFileHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFormatter(
            coloredlogs.ColoredFormatter(format)
        )


class ColoredStreamHandler(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFormatter(
            coloredlogs.ColoredFormatter(format)
        )

def _ignore(*args, **kwargs):
    return


class MockLogger:
    def __getattr__(self, item):
        return _ignore


def get_logger(name=''):
    if _LOG_LVL == 0:
        return MockLogger()

    filedir = "{}/logs/{}".format(dirname(cilantro_ee.__path__[0]), os.getenv('TEST_NAME', 'test'))
    filename = "{}/{}.log".format(filedir, os.getenv('HOST_NAME', name))

    if not os.path.exists(filedir):
        os.makedirs(filedir, exist_ok=True)

    filehandlers = [
        ColoredFileHandler('{}_color'.format(filename), delay=True, mode='a', maxBytes=5*1024*1024, backupCount=5),
        ColoredStreamHandler()
    ]

    logging.basicConfig(
        format=format,
        handlers=filehandlers,
        level=logging.DEBUG
    )

    log = logging.getLogger(name)
    log.setLevel(_LOG_LVL)

    for log_name, log_level in CUSTOM_LEVELS.items():
        apply_custom_level(log, log_name, log_level)

    return log


def overwrite_logger_level(level):
    global _LOG_LVL
    _LOG_LVL = level

    for name in logging.Logger.manager.loggerDict.keys():
        log = logging.getLogger(name)
        log.setLevel(level)



"""
BlockMetaData is just a subclass of BlockData, with no additional fields, so tests should be covered by test_block_data

-- davis
"""


# from cilantro_ee.messages.block_data.state_update import BlockDataReply
# from cilantro_ee.messages.block_data.block_data import BlockDataBuilder
# from cilantro_ee.constants.system_config import NUM_SUB_BLOCKS
# from unittest import TestCase
# import secrets
# from unittest import mock


# BlockDataReply is a subclass of BlockData, with no additions, so test_block_data should cover all the bases here


# class TestBlockDataReply(TestCase):
#
#     def test_init(self):
#         fbmds = [BlockDataBuilder.create_block(blk_num=1, sub_block_count=1) for _ in range(4)]
#
#         sr = BlockDataReply.create(fbmds)
#
#         self.assertEqual(fbmds, sr.block_data)
#
#     def test_serialization(self):
#         """
#         Tests serialize and from_bytes are inverse operations
#         """
#         fbmds = [BlockDataBuilder.create_block(blk_num=1, sub_block_count=1) for _ in range(4)]
#
#         sr = BlockDataReply.create(fbmds)
#         sr_bin = sr.serialize()
#
#         sr_clone = BlockDataReply.from_bytes(sr_bin)
#
#         self.assertEqual(sr, sr_clone)



from cilantro_ee.core.logger import get_logger
from cilantro_ee.constants import conf
from cilantro_ee.contracts import sync
from cilantro_ee.utils.test.testnet_config import read_public_constitution
from contracting.client import ContractingClient
import math

log = get_logger("VKBook")


class VKBook:
    def __init__(self, masternodes, delegates, num_boot_mns=1,
                 num_boot_del=1, stamps=True, nonces=False, debug=True):
        self.client = ContractingClient()

        self.contract = self.client.get_contract('vkbook')

        if self.contract is None or debug:
            # Put VKs into VKBook smart contract and submit it to state
            sync.submit_contract_with_construction_args('vkbook', args={'masternodes': masternodes,
                                                                        'delegates': delegates,
                                                                        'num_boot_mns': num_boot_mns,
                                                                        'num_boot_del': num_boot_del,
                                                                        'stamps': stamps,
                                                                        'nonces': nonces})

            self.contract = self.client.get_contract('vkbook')

        self.masternode_quorum_max = math.ceil(len(self.masternodes) * 2 / 3)

        self.delegate_quorum_max = math.ceil(len(self.delegates) * 2 / 3)

        num_boot_mns = self.contract.get_num_boot_masternodes()

        self.masternode_quorum_min = min(self.masternode_quorum_max, num_boot_mns)

        num_boot_del = self.contract.get_num_boot_delegates()

        self.delegate_quorum_min = min(self.delegate_quorum_max, num_boot_del)

        self.quorum_min = self.masternode_quorum_min + self.delegate_quorum_min


    @property
    def stamps_enabled(self):
        return self.contract.get_stamps_enabled()

    @property
    def nonces_enabled(self):
        return self.contract.get_nonces_enabled()

    @property
    def masternodes(self):
        return self.contract.get_masternodes()

    @property
    def delegates(self):
        return self.contract.get_delegates()

    @property
    def witnesses(self):
        return self.contract.get_witnesses()

    @property
    def notifiers(self):
        return self.contract.get_notifiers()

    @property
    def schedulers(self):
        return self.contract.get_schedulers()

    @property
    def state_sync(self):
        return self.masternodes + self.delegates

    @property
    def all(self):
        return self.masternodes + self.delegates + self.witnesses

    @property
    def num_boot_masternodes(self):
        return self.contract.get_num_boot_masternodes()

    @property
    def num_boot_delegates(self):
        return self.contract.get_num_boot_delegates()


book = read_public_constitution(conf.CONSTITUTION_FILE)
masternodes = [node['vk'] for node in book['masternodes']]
delegates = [node['vk'] for node in book['delegates']]

# print('masternodes before Phonebook init: {}'.format(masternodes))
# print('delegates before Phonebook init: {}'.format(delegates))

PhoneBook = VKBook(masternodes=masternodes,
                   delegates=delegates,
                   num_boot_mns=len(conf.BOOT_MASTERNODE_IP_LIST),
                   num_boot_del=len(conf.BOOT_DELEGATE_IP_LIST),
                   stamps=conf.STAMPS_ENABLED,
                   nonces=conf.NONCE_ENABLED)




from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.utils.test.dumpatron import Dumpatron
from cilantro_ee.utils.test.god import God
from cilantro_ee.utils.test.wallets import *
import time, random, os
from collections import defaultdict


CONTRACT_FILENAME = 'stubucks.sen.py'


class SmartContractTester(Dumpatron):

    ASSERT_CONTRACT_NOT_EXISTS_TIMEOUT = 6
    ASSERT_TX_TIMEOUT = 40
    ASSERT_SUBMITTED_TIMEOUT = 40
    ASSERT_SEEDED_TIMEOUT = 40
    SLEEP_BEFORE_ASSERTING = 10  # How long we should wait between sending transactions and asserting _new balances
    POLL_INTERVAL = 2

    TEST = 10

    CONTRACT_NAME = 'stubucks'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("SmartContractTester")

        self.sk1, self.vk1 = STU
        self.sk2, self.vk2 = DAVIS

    async def _start(self):
        self.assert_contract_does_not_exist()
        self.submit_contract()
        self.log.test("Sleeping for {} seconds before asserting contract was submitted...".format(self.SLEEP_BEFORE_ASSERTING))
        time.sleep(self.SLEEP_BEFORE_ASSERTING)
        self.assert_contract_submitted()
        self.assert_contract_seeded()

        self.submit_balance_tx()
        self.log.test("Sleeping for {} seconds before asserting balances tx work submitted...".format(self.SLEEP_BEFORE_ASSERTING))
        time.sleep(self.SLEEP_BEFORE_ASSERTING)
        self.assert_balance_tx_worked()
        self.submit_custodial_tx()
        self.log.test("Sleeping for {} seconds before asserting custodial tx work submitted...".format(self.SLEEP_BEFORE_ASSERTING))
        time.sleep(self.SLEEP_BEFORE_ASSERTING)
        self.assert_custodial_tx_worked()

    def assert_contract_does_not_exist(self):
        names = God.get_contract_names()
        if self.CONTRACT_NAME in names:
            raise Exception("Contract named '{}' already exists in Masternode's contracts {}".format(self.CONTRACT_NAME, names))

    def submit_contract(self):
        self.log.test("Submitting smart contract with owner vk {}".format(self.vk1))
        code_file_path = os.path.dirname(__file__) + '/' + CONTRACT_FILENAME
        with open(code_file_path, 'r') as f:
            code_str = f.read()
        God.submit_contract(code_str, self.CONTRACT_NAME, self.sk1, self.vk1, stamps=self.STAMPS_AMOUNT)

    def assert_contract_submitted(self):
        self.log.test("Asserting that contract has been successfully submitted...")
        elapsed = 0
        while self.CONTRACT_NAME not in God.get_contract_names():
            self.log.info("Contract still not registered. Sleeping {} seconds".format(self.POLL_INTERVAL))
            time.sleep(self.POLL_INTERVAL)
            elapsed += self.POLL_INTERVAL
            if elapsed > self.ASSERT_SUBMITTED_TIMEOUT:
                raise Exception("Contract submission exceeded timeout of {}".format(self.ASSERT_SUBMITTED_TIMEOUT))

        self.log.test("Contract successfully submitted.")

    def assert_contract_seeded(self):
        self.log.test("Asserting that contract has been successfully seeded...")
        elapsed = 0
        while not self._check_value('balances', self.vk1, 1000000):
            time.sleep(self.POLL_INTERVAL)
            elapsed += self.POLL_INTERVAL
            if elapsed > self.ASSERT_SEEDED_TIMEOUT:
                raise Exception("Contract seeding exceeded timeout of {}".format(self.ASSERT_SEEDED_TIMEOUT))

    def submit_balance_tx(self):
        tx = God.create_contract_tx(self.sk1, self.CONTRACT_NAME, 'transfer', self.STAMPS_AMOUNT, to=self.vk2,
                                    amount=1337)
        self.log.test("Sending balance transaction {}".format(tx))
        God.send_tx(tx)

    def submit_custodial_tx(self):
        tx = God.create_contract_tx(self.sk2, self.CONTRACT_NAME, 'add_to_custodial', self.STAMPS_AMOUNT, to=self.vk1,
                                    amount=337)
        self.log.test("Sending custodial transaction {}".format(tx))
        God.send_tx(tx)

    def _get_sc_data(self, table, key):
        return God.get_from_mn_api('contracts/{}/{}/{}'.format(self.CONTRACT_NAME, table, key))

    def _check_value(self, table, key, expected_value, assume_int=True) -> bool:
        full_key = "{}/{}/{}".format(self.CONTRACT_NAME, table, key)
        actual = self._get_sc_data(table, key)

        if actual is None:
            actual = '<fetch returned None>'
        else:
            if assume_int:
                assert 'value' in actual, "Return JSON for key {} expected to have key 'value', but instead {}"\
                                          .format(full_key, actual)
                if actual['value'] == 'null':
                    actual = '<fetch returned None>'
                else:
                    actual = int(actual['value'])

        if actual != expected_value:
            self.log.warning("Key {} expected to have value {} but actually has value {}"
                             .format(full_key, expected_value, actual))
            return False
        return True

    def assert_balance_tx_worked(self):
        def _check_sender_deducted():
            return self._check_value('balances', self.vk1, 1000000 - 1337)

        def _check_recv_funded():
            return self._check_value('balances', self.vk2, 1337)

        elapsed = 0
        self.log.test("Checking to make sure stubucks were sent...")
        while not (_check_sender_deducted() and _check_recv_funded()):
            elapsed += self.POLL_INTERVAL
            time.sleep(self.POLL_INTERVAL)
            if elapsed > self.ASSERT_TX_TIMEOUT:
                raise Exception("Exceeded timeout of {} waiting for balance assertions to pass"
                                .format(self.ASSERT_TX_TIMEOUT))

    def assert_custodial_tx_worked(self):
        def _check_sender_deducted():
            return self._check_value('balances', self.vk2, 1337 - 337)

        def _check_recv_funded():
            return self._check_value('custodials', self.vk1, 337)

        elapsed = 0
        self.log.test("Checking to make sure custodials were sent...")
        while not (_check_sender_deducted() and _check_recv_funded()):
            elapsed += self.POLL_INTERVAL
            time.sleep(self.POLL_INTERVAL)
            if elapsed > self.ASSERT_TX_TIMEOUT:
                raise Exception("Exceeded timeout of {} waiting for custodial assertions to pass"
                                .format(self.ASSERT_TX_TIMEOUT))





from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.utils.test.god import God
from cilantro_ee.constants.system_config import *
import os, glob
import asyncio, aiohttp


SSL_ENABLED = False  # TODO make this infered instead of a hard coded flag

IP_FILE_PREFIX = 'ip_masternode'


class Dumpatron:

    TX_PER_BLOCK = 0
    STAMPS_AMOUNT = 0

    def __init__(self, env_path):
        if env_path[-1] != '/':
            env_path += '/'
        assert os.path.exists(env_path), "No env dir found at path {}".format(env_path)
        assert os.path.exists(env_path + '.cache'), "No .cache dir found at path {}.cache".format(env_path)
        self.log = get_logger("Dumpatron")
        Dumpatron.TX_PER_BLOCK = TRANSACTIONS_PER_SUB_BLOCK * NUM_SB_PER_BLOCK * NUM_BLOCKS

        self.env_path = env_path
        self.mn_ip_dict = self._extract_mn_ips()
        self.mn_url_list = self._get_mn_urls_from_ips(self.mn_ip_dict)

        God.mn_urls = self.mn_url_list
        God.multi_master = True

        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.session = aiohttp.ClientSession()

        God.wait_for_mns_online()

    async def _start(self):
        await self.start_interactive_dump()

    async def _start_loop(self):
        try:
            await self._start()
        except asyncio.CancelledError:
            self.log.warning("Loop cancelled")
        finally:
            await self.session.close()

    def start(self):
        self.loop.run_until_complete(self._start_loop())

    def _extract_mn_ips(self) -> dict:
        ips = {}
        pattern = "{}.cache/{}*".format(self.env_path, IP_FILE_PREFIX)
        mn_files = glob.glob(pattern)
        assert len(mn_files) > 0, "No masternode ip config files found matching glob pattern {} (Did you do make run " \
                                  "first? Otherwise, harass Colin...did he change the" \
                                  " way IPs are cached?)".format(pattern)

        for mn_file in mn_files:
            with open(mn_file, 'r') as f:
                mn_idx = mn_file[len(IP_FILE_PREFIX):]
                ips[mn_file[-1]] = f.read()
        return ips

    # TODO factor this out into its own function/module so it can be used in God as well
    def _get_mn_urls_from_ips(self, ips: dict) -> list:
        urls = []
        for ip in ips.values():
            if SSL_ENABLED:
                urls.append("https://{}".format(ip))
            else:
                urls.append("http://{}:8080".format(ip))
        return urls

    async def start_interactive_dump(self):
        self.log.info("Starting the dump....")
        while True:
            user_input = input("Enter an integer representing the # of transactions to dump, or 'x' to quit. "
                               "Press enter to dump 1 blocks\n")

            if user_input.lower() == 'x':
                self.log.important("Termination input detected. Breaking")
                break

            vol = int(user_input) if user_input.isdigit() else self.TX_PER_BLOCK
            self.log.important3("Dumping {} transactions!".format(vol))
            await God.dump_it(self.session, volume=vol)

    def _parse_reply(self, req, req_type='json'):
        if req.status_code != 200:
            self.log.spam("Got status code {} from request {}".format(req.status_code, req))
            return None

        if req_type == 'json':
            return req.json()
        else:
            raise Exception("Unknown request type {}".format(req_type))






import hashlib

class BlockNotification:
    # raghu todo - this function should go into utils or hash utils?
    @staticmethod
    def get_block_hash(prev_block_hash, input_hashes):
        if type(prev_block_hash) == str:
            prev_block_hash = bytes.fromhex(prev_block_hash)

        h = hashlib.sha3_256()
        h.update(prev_block_hash)
        for ihes in input_hashes:
            for ih in ihes:
                h.update(bytes.fromhex(ih) if type(ih) == str else ih)
        return h.digest()

    # todo - needs more details on which sbs' are used to make partial block
    @staticmethod
    def get_partial_block_notification(block_num, block_hash, block_owners, first_sb_idx, input_hashes):
        # bn = BlockNotification.get_block_notification(block_num, block_hash,
                      # block_owners, first_sb_idx, [[ih] for ih in input_hashes])
        # bn.type.partialBlock = None
        # return bn
        pass





from unittest import TestCase
from cilantro_ee.nodes.masternode.block_contender import SubBlockGroup, BlockContender
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.services.storage.vkbook import VKBook
from cilantro_ee.core.crypto.wallet import Wallet
import secrets
from tests import random_txs
from collections import namedtuple
import hashlib
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
from cilantro_ee.constants.system_config import NUM_SB_PER_BLOCK, NUM_SB_BUILDERS
import os
import capnp

transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')

def random_wallets(n=10):
    return [secrets.token_hex(32) for _ in range(n)]


class TestSubBlockGroup(TestCase):
    def setUp(self):
        m = MetaDataStorage()
        m.flush()

    def test_init(self):
        SubBlockGroup(0, 'A' * 64)

    def test_consensus_not_reached_after_init(self):
        contacts = VKBook(delegates=['A' * 64], masternodes=['A' * 64])
        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        self.assertFalse(s.is_consensus_reached())

    def test_current_quorum_reached_is_zero_if_best_result_has_less_than_min_quorum(self):
        contacts = VKBook(delegates=['A' * 64, 'B' * 64, 'C' * 64], masternodes=['A' * 64],
                          num_boot_del=3)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64

        s.rh[s.best_rh] = {'A' * 64}

        self.assertEqual(s.get_current_quorum_reached(), 0)

    def test_current_quorum_reached_is_max_quorum_if_best_result_has_max_quorum_votes(self):
        contacts = VKBook(delegates=['A' * 64, 'B' * 64, 'C' * 64], masternodes=['A' * 64],
                          num_boot_del=3)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64

        s.rh[s.best_rh] = {'A' * 64, 'B' * 64, 'C' * 64}

        self.assertEqual(s.get_current_quorum_reached(), s.max_quorum)

    def test_current_quorum_reached_is_max_quorum_if_best_result_has_more_than_max_quorum_votes(self):
        contacts = VKBook(delegates=['A' * 64, 'B' * 64, 'C' * 64, 'D' * 64], masternodes=['A' * 64],
                          num_boot_del=3)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64

        s.rh[s.best_rh] = {'A' * 64, 'B' * 64, 'C' * 64, 'D' * 64}

        self.assertEqual(s.get_current_quorum_reached(), s.max_quorum)

    def test_current_quorum_returns_zero_when_no_result_has_enough_votes(self):
        contacts = VKBook(delegates=['A' * 64, 'B' * 64, 'C' * 64, 'D' * 64], masternodes=['A' * 64],
                          num_boot_del=3)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64
        another_result = 'C' * 64

        s.rh[s.best_rh] = {'A' * 64}
        s.rh[another_result] = {'B' * 64}

        self.assertEqual(s.get_current_quorum_reached(), 0)

    def test_current_quorum_returns_leading_vote_number_if_reduced_quorum_can_be_set(self):
        delegates = random_wallets(100)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64
        another_result = 'C' * 64

        s.rh[s.best_rh] = set(delegates[:59])
        s.rh[another_result] = {'B' * 64}

        self.assertEqual(s.get_current_quorum_reached(), 59)

    def test_consensus_is_possible_returns_false_when_no_result_has_quorum(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64}
        s.rh[2] = {'B' * 64}
        s.rh[3] = {'B' * 64}
        s.rh[4] = {'B' * 64}
        s.rh[5] = {'B' * 64}
        s.rh[6] = {'B' * 64}
        s.rh[7] = {'B' * 64}
        s.rh[8] = {'B' * 64}
        s.rh[9] = {'B' * 64}
        s.rh[0] = {'B' * 64}

        self.assertFalse(s.is_consensus_possible())

    def test_consensus_is_possible_returns_true_when_a_single_result_has_max_quorum(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}
        s.rh[2] = {'B' * 64}

        self.assertTrue(s.is_consensus_possible())

    def test_consensus_is_possible_if_there_are_still_enough_votes_left_to_make_quorum(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64}
        s.rh[2] = {'B' * 64, 'C' * 64, 'D' * 64}

        self.assertTrue(s.is_consensus_possible())

    def test_get_input_hashes_returns_list_of_hashes_in_group(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        SBCInput = namedtuple('SBCInput', ['inputHash'])

        s.sender_to_sbc = {
            delegates[0]: SBCInput(b'A' * 32),
            delegates[1]: SBCInput(b'B' * 32),
            delegates[2]: SBCInput(b'C' * 32),
            delegates[3]: SBCInput(b'D' * 32),
            delegates[4]: SBCInput(b'E' * 32),
        }

        expected = [b'A' * 32, b'B' * 32, b'C' * 32, b'D' * 32, b'E' * 32]

        self.assertEqual(set(s.get_input_hashes()), set(expected))

    def test_get_merkle_leaves_returns_empty_list_if_no_best_rh(self):
        pass

    # *** #

    def test_verify_sbc_false_sender_ne_merkle_proof_signer(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        self.assertFalse(s._verify_sbc(sender_vk=Wallet().verifying_key(), sbc=sbc))

    def test_verify_sbc_false_sbc_idx_ne_self_sb_idx(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=2)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_false_invalid_sig(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=0, poisoned_sig=b'\x00' * 64)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_false_prev_block_hash_ne_curr_block_hash(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, b'B' * 32, w=sender)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_false_sbc_merkle_leave_does_not_verify(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, poison_result_hash=True)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_false_tx_hash_not_in_merkle_leaves(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, poison_tx=True)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_false_sb_idx_gte_num_sb_per_block(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(200, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=200)

        self.assertFalse(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_verify_sbc_true_if_no_failures(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        self.assertTrue(s._verify_sbc(sender_vk=sender.verifying_key(), sbc=sbc))

    def test_get_sbc_builds_best_sb_and_returns(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.best_rh = sbc.resultHash
        s.rh[s.best_rh] = {sbc}

        got_sbc = s.get_sb()

        self.assertEqual(sbc.resultHash, got_sbc.merkleRoot)
        self.assertEqual(sbc.subBlockIdx, got_sbc.subBlockIdx)
        self.assertEqual(sbc.inputHash, got_sbc.inputHash)
        self.assertEqual(sbc.signature, got_sbc.signatures[0])
        self.assertListEqual([leaf for leaf in sbc.merkleLeaves],
                             [leaf for leaf in got_sbc.merkleLeaves])

    def test_get_ordered_transactions_returns_properly(self):
        SBCMockTXs = namedtuple('SBCMockTXs', ['transactions'])

        best_sbc = SBCMockTXs(transactions=(1, 2, 3, 4, 5))

        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        s.best_rh = b'B' * 32
        s.rh[s.best_rh] = {best_sbc}

        txs = s._get_ordered_transactions()

        self.assertEqual((1, 2, 3, 4, 5), txs)

    def test_get_merkle_leaves_returns_properly(self):
        SBCMockMerkle = namedtuple('SBCMockMerkle', ['merkleLeaves'])

        best_sbc = SBCMockMerkle(merkleLeaves=(1, 2, 3, 4, 5))

        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        s.best_rh = b'B' * 32
        s.rh[s.best_rh] = {best_sbc}

        txs = s._get_merkle_leaves()

        self.assertEqual((1, 2, 3, 4, 5), txs)

    def test_get_merkle_leaves_returns_empty_list_if_no_best_rh(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        self.assertListEqual(s._get_merkle_leaves(), [])

    def test_is_empty_returns_true_if_no_merkle_leaves(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        self.assertTrue(s.is_empty())

    def test_is_empty_returns_false_if_merkle_leaves(self):
        SBCMockMerkle = namedtuple('SBCMockMerkle', ['merkleLeaves'])

        best_sbc = SBCMockMerkle(merkleLeaves=(1, 2, 3, 4, 5))

        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        s.best_rh = b'B' * 32
        s.rh[s.best_rh] = {best_sbc}

        self.assertFalse(s.is_empty())

    def test_add_sbc_returns_none_if_cannot_verify(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, poison_tx=True)

        self.assertFalse(s.add_sbc(sender.verifying_key(), sbc))

    def test_add_sbc_sender_already_exists_overwrites_with_new_sbc(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)
        sbc_2 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        self.assertEqual(s.sender_to_sbc[sender.verifying_key()], sbc_1)

        s.add_sbc(sender.verifying_key(), sbc_2)

        self.assertEqual(s.sender_to_sbc[sender.verifying_key()], sbc_2)

    def test_add_sbc_adds_map_of_sender_vk_to_sbc(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        self.assertEqual(s.sender_to_sbc[sender.verifying_key()], sbc_1)

    def test_add_sbc_adds_to_result_hash_set(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        self.assertEqual(s.rh[sbc_1.resultHash], {sbc_1})

# (self.best_rh is None) or (len(self.rh[sbc.resultHash]) > len(self.rh[self.best_rh]))

    def test_add_sbc_best_rh_none_sets_best_to_submitted(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        self.assertEqual(s.best_rh, sbc_1.resultHash)

    def test_add_sbc_added_result_hash_set_gt_current_best_rh_sets_best_rh_to_submitted(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender_1 = Wallet()
        sender_2 = Wallet()

        sbc_1, sbc_2 = random_txs.double_sbc_from_tx(input_hash, s.curr_block_hash, w1=sender_1, w2=sender_2)

        sender_3 = Wallet()

        sbc_3 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender_3)

        s.add_sbc(sender_3.verifying_key(), sbc_3)

        self.assertEqual(s.best_rh, sbc_3.resultHash)

        s.add_sbc(sender_2.verifying_key(), sbc_2)

        self.assertEqual(s.best_rh, sbc_3.resultHash)

        s.add_sbc(sender_1.verifying_key(), sbc_1)

        self.assertEqual(s.best_rh, sbc_2.resultHash)

    def test_tx_hashes_added_to_transaction_hash_to_tx_object(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        self.assertListEqual(list(s.transactions.values()), [tx for tx in sbc_1.transactions])

        for tx in sbc_1.transactions:
            h = hashlib.sha3_256()
            h.update(tx)
            _hash = h.digest()

            self.assertEqual(s.transactions[_hash], tx)


class TestBlockContender(TestCase):
    def setUp(self):
        m = MetaDataStorage()
        m.flush()

    def test_reset_resets_all_state(self):
        b = BlockContender()

        b.committed = True
        b.consensus_reached = True
        b.sb_groups = {1, 2, 3}

        m = MetaDataStorage()

        m.set_latest_block_hash(b'C' * 32)

        b.reset()

        self.assertEqual(b.committed, False)
        self.assertEqual(b.consensus_reached, False)
        self.assertEqual(b.sb_groups, {})
        self.assertEqual(b.curr_block_hash, b'C' * 32)

    def test_is_consensus_reached_false_if_len_sb_groups_lt_num_sb_per_block(self):
        b = BlockContender()

        b.sb_groups = {1, 2}

        self.assertFalse(b.is_consensus_reached())

    def test_consensus_not_reached_on_any_sb_group_returns_false(self):
        contacts = VKBook(delegates=['A' * 64], masternodes=['A' * 64])
        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)
        t = SubBlockGroup(1, 'A' * 64, contacts=contacts)
        u = SubBlockGroup(2, 'A' * 64, contacts=contacts)

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertFalse(b.is_consensus_reached())

    def test_consensus_reached_returns_true_if_all_sb_groups_are_reached(self):
        contacts = VKBook(delegates=['A' * 64, 'B' * 64, 'C' * 64, 'D' * 64], masternodes=['A' * 64],
                          num_boot_del=3)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 'B' * 64

        s.rh[s.best_rh] = {'A' * 64, 'B' * 64, 'C' * 64, 'D' * 64}

        t = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        t.best_rh = 'B' * 64

        t.rh[s.best_rh] = {'A' * 64, 'B' * 64, 'C' * 64, 'D' * 64}

        u = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        u.best_rh = 'B' * 64

        u.rh[s.best_rh] = {'A' * 64, 'B' * 64, 'C' * 64, 'D' * 64}

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertTrue(b.is_consensus_reached())

    def test_consensus_not_possible_if_any_sb_groups_not_possible_returns_false(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64}
        s.rh[2] = {'B' * 64}
        s.rh[3] = {'B' * 64}
        s.rh[4] = {'B' * 64}
        s.rh[5] = {'B' * 64}
        s.rh[6] = {'B' * 64}
        s.rh[7] = {'B' * 64}
        s.rh[8] = {'B' * 64}
        s.rh[9] = {'B' * 64}
        s.rh[0] = {'B' * 64}

        b = BlockContender()
        b.sb_groups = {
            0: s,
        }

        self.assertFalse(b.is_consensus_possible())

    def test_consensus_possible_if_all_sb_groups_are_possible_returns_true(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        t = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        t.best_rh = 1
        t.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        u = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        u.best_rh = 1
        u.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertTrue(b.is_consensus_possible())

    def test_get_current_quorum_reached_returns_0_if_not_enough_sb_per_block(self):
        b = BlockContender()
        b.sb_groups = {
            0: 1,
        }

        self.assertEqual(b.get_current_quorum_reached(), 0)

    def test_get_current_quorum_reached_returns_largest_quorum_if_less_than_max_quorum(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        s.best_rh = 1
        s.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        t = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        t.best_rh = 1
        t.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        u = SubBlockGroup(0, 'A' * 64, contacts=contacts)

        u.best_rh = 1
        u.rh[1] = {'B' * 64, 'C' * 64, 'D' * 64, 'E' * 64, 'F' * 64, '0' * 64, '1' * 64}

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertEqual(b.get_current_quorum_reached(), 2)

    def test_is_empty_returns_false_if_any_sb_in_group_is_not_empty(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)
        SBCMockMerkle = namedtuple('SBCMockMerkle', ['merkleLeaves'])

        best_sbc = SBCMockMerkle(merkleLeaves=(1, 2, 3, 4, 5))

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)
        s.best_rh = b'B' * 32
        s.rh[s.best_rh] = {best_sbc}

        t = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        u = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertFalse(b.is_empty())

    def test_is_empty_returns_true_if_all_sb_in_group_empty(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)
        SBCMockMerkle = namedtuple('SBCMockMerkle', ['merkleLeaves'])

        best_sbc = SBCMockMerkle(merkleLeaves=(1, 2, 3, 4, 5))

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        t = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        u = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        self.assertTrue(b.is_empty())

    def test_get_sb_data_returns_sorted_list_by_subBlockIdx(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        t = SubBlockGroup(1, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_2 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=1)

        t.add_sbc(sender.verifying_key(), sbc_2)

        u = SubBlockGroup(2, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_3 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=2)

        u.add_sbc(sender.verifying_key(), sbc_3)

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        sb_data = b.get_sb_data()

        self.assertEqual(sbc_1.resultHash, sb_data[0].merkleRoot)
        self.assertEqual(sbc_1.subBlockIdx, sb_data[0].subBlockIdx)

        self.assertEqual(sbc_2.resultHash, sb_data[1].merkleRoot)
        self.assertEqual(sbc_2.subBlockIdx, sb_data[1].subBlockIdx)

        self.assertEqual(sbc_3.resultHash, sb_data[2].merkleRoot)
        self.assertEqual(sbc_3.subBlockIdx, sb_data[2].subBlockIdx)

    def test_add_sbc_returns_false_if_hash_already_in_old_input_hashes(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        b = BlockContender()
        b.old_input_hashes.add(input_hash)

        self.assertFalse(b.add_sbc(sender.verifying_key(), sbc))

    def test_add_sbc_adds_subblock_group_if_sb_idx_doesnt_exist(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        b = BlockContender()

        self.assertIsNone(b.sb_groups.get(0))

        b.add_sbc(sender.verifying_key(), sbc)

        sb_group = b.sb_groups.get(0)

        self.assertEqual(sb_group.sb_idx, 0)
        self.assertEqual(sb_group.curr_block_hash, b.curr_block_hash)

    def test_add_sbc_adds_sbc_to_proper_index(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        b = BlockContender()
        b.add_sbc(sender.verifying_key(), sbc)

        got_sb = b.sb_groups[0].get_sb()

        self.assertEqual(sbc.resultHash, got_sb.merkleRoot)
        self.assertEqual(sbc.subBlockIdx, got_sb.subBlockIdx)

        s = SubBlockGroup(1, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=1)

        b.add_sbc(sender.verifying_key(), sbc)

        got_sb = b.sb_groups[1].get_sb()

        self.assertEqual(sbc.resultHash, got_sb.merkleRoot)
        self.assertEqual(sbc.subBlockIdx, got_sb.subBlockIdx)

    def test_add_sbc_returns_true_if_no_groups(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        b = BlockContender()

        self.assertTrue(b.add_sbc(sender.verifying_key(), sbc))

    def test_add_sbc_returns_false_if_groups_already_exist(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)

        sender = Wallet()

        sbc = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        b = BlockContender()
        b.sb_groups[0] = s

        self.assertFalse(b.add_sbc(sender.verifying_key(), sbc))

    def test_get_first_sb_idx_unsorted_returns_properly(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        b = BlockContender()
        b.sb_groups[0] = s

        self.assertEqual(b.get_first_sb_idx_unsorted(b.sb_groups), 0)

    def test_get_first_sb_idx_sorted_returns_properly(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'\x00' * 32, contacts=contacts)

        b = BlockContender()
        b.sb_groups[0] = s

        self.assertEqual(b.get_first_sb_idx_sorted(), 0)

    def test_get_input_hashes_sorted_works(self):
        delegates = random_wallets(10)

        contacts = VKBook(delegates=delegates,
                          masternodes=['A' * 64],
                          num_boot_del=10)

        s = SubBlockGroup(0, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_1 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender)

        s.add_sbc(sender.verifying_key(), sbc_1)

        t = SubBlockGroup(1, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_2 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=1)

        t.add_sbc(sender.verifying_key(), sbc_2)

        u = SubBlockGroup(2, b'A' * 32, contacts=contacts)

        input_hash = secrets.token_bytes(32)
        sender = Wallet()
        sbc_3 = random_txs.sbc_from_txs(input_hash, s.curr_block_hash, w=sender, idx=2)

        u.add_sbc(sender.verifying_key(), sbc_3)

        b = BlockContender()
        b.sb_groups = {
            0: s,
            1: t,
            2: u
        }

        input_hashes = b.get_input_hashes_sorted()

        self.assertEqual(input_hashes[0][0], sbc_1.inputHash)
        self.assertEqual(input_hashes[1][0], sbc_2.inputHash)
        self.assertEqual(input_hashes[2][0], sbc_3.inputHash)


from unittest import TestCase
from cilantro_ee.core import canonical
from tests import random_txs


class TestCanonicalCoding(TestCase):
    def test_recursive_dictionary_sort_works(self):
        unsorted = {
            'z': 123,
            'a': {
                'z': 123,
                'a': 532,
                'x': {
                    'a': 123,
                    'vvv': 54
                }
            }
        }

        sorted_dict = {
            'a': {
                'a': 532,
                'x': {
                 'a': 123,
                 'vvv': 54
                },
                'z': 123,
            },
            'z': 123
        }

        s = canonical.format_dictionary(unsorted)

        self.assertDictEqual(s, sorted_dict)

    def test_random_subblock_converts_successfully(self):
        sb = random_txs.random_block().subBlocks[0].to_dict()

        expected_order = ['inputHash', 'merkleLeaves', 'merkleRoot', 'signatures', 'subBlockIdx', 'transactions']

        sorted_sb = canonical.format_dictionary(sb)
        sorted_sb_keys = list(sorted_sb.keys())
        for i in range(len(sorted_sb_keys)):
            self.assertEqual(sorted_sb_keys[i], expected_order[i])

    def test_random_subblocks_all_convert_successfully_at_top_level(self):
        expected_order = ['inputHash', 'merkleLeaves', 'merkleRoot', 'signatures', 'subBlockIdx', 'transactions']

        block = random_txs.random_block()
        sbs = [block.subBlocks[i].to_dict() for i in range(len(block.subBlocks))]

        for sb in sbs:
            sorted_sb = canonical.format_dictionary(sb)
            sorted_sb_keys = list(sorted_sb.keys())
            for i in range(len(sorted_sb_keys)):
                self.assertEqual(sorted_sb_keys[i], expected_order[i])

    def test_random_subblock_all_transactiondata_convert_successfully(self):
        sb = random_txs.random_block().subBlocks[0].to_dict()

        sb = canonical.format_dictionary(sb)

        expected_order = ['stampsUsed', 'state', 'status', 'transaction']

        for tx in sb['transactions']:
            sorted_tx_keys = list(tx.keys())
            self.assertEqual(sorted_tx_keys, expected_order)

    def test_random_subblock_all_transactions_convert_successfully(self):
        sb = random_txs.random_block().subBlocks[0].to_dict()

        expected_order = ['contractName', 'functionName', 'kwargs', 'nonce', 'processor', 'sender', 'stampsSupplied']

        sb = canonical.format_dictionary(sb)

        for tx in sb['transactions']:
            sorted_tx_keys = list(tx['transaction']['payload'].keys())
            self.assertEqual(sorted_tx_keys, expected_order)

    def test_block_from_subblocks_verify_works(self):
        sbs = random_txs.random_block().subBlocks

        block = canonical.block_from_subblocks(subblocks=sbs)

        prev_hash = block['prevBlockHash']
        prop_hash = block['blockHash']
        subblocks = block['subBlocks']

        valid = canonical.verify_block(subblocks, prev_hash, prop_hash)

        self.assertTrue(valid)




import time
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.capnp_impl.capnp_impl import CapnpImpl
from cilantro_ee.core.crypto.wallet import Wallet
import struct

class Message:

    _msg_impl = CapnpImpl()   # currently only capnp implementation is available


    # returns msg_type, encoded_msg
    @classmethod
    def get_message(cls, msg_type: MessageType, **kwargs):
        return cls._msg_impl.get_message(msg_type=msg_type, **kwargs)

    # returns msg_type, encoded_msg_packed
    @classmethod
    def get_message_packed(cls, msg_type: MessageType, **kwargs):
        return cls._msg_impl.get_message_packed(msg_type=msg_type, **kwargs)

    # returns encoded_msg_type, signed_encoded_msg_packed
    @classmethod
    def get_signed_message_packed(cls, wallet: Wallet, msg_type: MessageType, **kwargs):
        return cls._msg_impl.get_signed_message_packed(
                                  wallet=wallet,
                                  msg_type=msg_type,**kwargs)

    # returns encoded_msg_type, signed_encoded_msg_packed
    @classmethod
    def get_message_signed_internal(cls, signee: bytes, sign: callable, msg_type: bytes, msg: bytes):
        return cls._msg_impl._get_message_signed(
                                  signee=signee, sign=sign,
                                  msg_type=msg_type, msg=msg)

    # returns encoded_msg

    @classmethod
    def unpack_message_internal(cls, msg_type: MessageType, message: bytes):
        _, msg, _, _, _ = cls._msg_impl._unpack_message(
            msg_type=msg_type, message=message,
            is_verify=False)

        return msg

    # returns msg_type, encoded_msg, sender, timestamp, is_verified
    @classmethod
    def unpack_message(cls, msg_type: bytes, message: bytes,
                       sender: bytes = None, timestamp: float = time.time(),
                       is_verify: bool = True):
        return cls._msg_impl.unpack_message(
                                 msg_type=msg_type, message=message,
                                 sender=sender, timestamp=timestamp,
                                 is_verify=is_verify)

    @classmethod
    def unpack_message_2(cls, message: bytes):
        msg_type = struct.pack('B', message[0])
        msg_blob = message[1:]

        return cls._msg_impl.unpack_message(msg_type=msg_type, message=msg_blob)

    @classmethod
    def get_signed_message_packed_2(cls, wallet: Wallet, msg_type: MessageType, **kwargs):
        t, m = cls._msg_impl.get_signed_message_packed(
            wallet=wallet,
            msg_type=msg_type, **kwargs)

        return t + m




from cilantro_ee.core.sockets.socket import SocketUtil
from cilantro_ee.core.utils.context import Context
from cilantro_ee.core.logger import get_logger
from cilantro_ee.services.overlay.server import OverlayServer
from cilantro_ee.utils.lprocess import LProcess
from cilantro_ee.services.storage.vkbook import VKBook, PhoneBook
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.constants import conf
import asyncio
import time


class NodeTypes:
    MN = 'mn'
    WITNESS = 'w'
    DELEGATE = 'd'

    _ALL_TYPES = {MN, WITNESS, DELEGATE}

    @classmethod
    def check_vk_in_group(cls, vk: str, group: str):
        assert group in cls._ALL_TYPES, "Group '{}' not a valid node type. Must be in {}".format(group, cls._ALL_TYPES)
        if group == cls.MN:
            return vk in PhoneBook.masternodes
        if group == cls.WITNESS:
            return vk in PhoneBook.witnesses
        if group == cls.DELEGATE:
            return vk in PhoneBook.delegates


class NodeBase(Context):

    def __init__(self, ip, signing_key, name='Node'):
        super().__init__(signing_key=signing_key, name=name)
        
        SocketUtil.clear_domain_register()

        self.log = get_logger(name)
        self.ip = ip
        self.wallet = Wallet(seed=signing_key)

        conf.HOST_VK = self.wallet.verifying_key()

        self.log.info("Starting node components")
        quorum = self.start_node()

        self.log.info("Starting overlay service")
        self.overlay_server = OverlayServer(sk=signing_key, ctx=self.zmq_ctx, quorum=1)
        self.start()

    def start(self):
        self.overlay_server.start()

    def start_node(self):
        pass




delegate_list = Variable()
masternode_list = Variable()
witness_list = Variable()
notifier_list = Variable()
scheduler_list = Variable()

num_boot_masternodes = Variable()
num_boot_delegates = Variable()

stamps_enabled = Variable()
nonces_enabled = Variable()



@construct
def seed(masternodes,
         delegates,
         num_boot_mns,
         num_boot_del,
         stamps=True,
         nonces=False):

    masternode_list.set(masternodes)
    delegate_list.set(delegates)
    witness_list.set([])
    notifier_list.set([])
    scheduler_list.set([])

    num_boot_masternodes.set(num_boot_mns)
    num_boot_delegates.set(num_boot_del)

    stamps_enabled.set(stamps)
    nonces_enabled.set(nonces)

@export
def get_delegates():
    return delegate_list.get()


@export
def get_masternodes():
    return masternode_list.get()


@export
def get_witnesses():
    return witness_list.get()

@export
def get_notifiers():
    return notifier_list.get()


@export
def get_schedulers():
    return scheduler_list.get()


@export
def get_num_boot_masternodes():
    return num_boot_masternodes.get()

@export
def get_num_boot_delegates():
    return num_boot_delegates.get()


@export
def get_stamps_enabled():
    return stamps_enabled.get()


@export
def get_nonces_enabled():
    return nonces_enabled.get()



from contracting.db.driver import ContractDriver

EPOCH_INTERVAL = 100
EPOCH_HASH_KEY = '__epoch_hash'
EPOCH_NUMBER_KEY = '__epoch_number'


class EpochManager:
    def __init__(self, driver=ContractDriver()):
        self.driver = driver

    def update_epoch_if_needed(self, block):
        if block.blockNum % EPOCH_INTERVAL == 0:
            self.set_epoch_hash(block.blockHash)
            self.set_epoch_number(block.blockNum // EPOCH_INTERVAL)

    def get_epoch_hash(self):
        return self.driver.get(EPOCH_HASH_KEY) or b'x/00' * 32

    def set_epoch_hash(self, value):
        self.driver.set(EPOCH_HASH_KEY, value)

    def get_epoch_number(self):
        return self.driver.get(EPOCH_NUMBER_KEY) or 0

    def set_epoch_number(self, value):
        self.driver.set(EPOCH_NUMBER_KEY, value)



from unittest import TestCase
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.utils.pow import SHA3POWBytes
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.capnp_impl import capnp_struct as schemas
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.capnp_impl.capnp_impl import CapnpImpl

import os
import capnp

blockdata_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/blockdata.capnp')


class TestMessages(TestCase):
    def test_message_new_api(self):
        capnp_impl = CapnpImpl()
        message = blockdata_capnp.BlockDataRequest.new_message(blockNum=123)
        msgType, same_message = capnp_impl.get_message(msg_type=MessageType.BLOCK_DATA_REQUEST, blockNum=123)

        mtype1, msg1 = Message.get_message(MessageType.BLOCK_DATA_REQUEST, blockNum=123)
       
        self.assertEqual(msgType, mtype1)
        self.assertEqual(message.blockNum, msg1.blockNum)

        self.assertEqual(message.blockNum, same_message.blockNum)

    def test_unpack_returns_equal_capnp_struct(self):
        capnp_impl = CapnpImpl()
        msgType, message = capnp_impl.get_message(msg_type=MessageType.BLOCK_DATA_REQUEST, blockNum=237)
        msgType2, packed = capnp_impl.get_message_packed(msg_type=MessageType.BLOCK_DATA_REQUEST, blockNum=237)

        unpacked_message = blockdata_capnp.BlockDataRequest.from_bytes_packed(packed)

        self.assertEqual(msgType, msgType2)
        self.assertEqual(message.blockNum, unpacked_message.blockNum)

    def test_signal(self):
        capnp_impl = CapnpImpl()
        msg_type, message = capnp_impl.get_message(msg_type=MessageType.READY)
        self.assertEqual(message, '')
        msg_type, message = capnp_impl.get_message_packed(msg_type=MessageType.MAKE_NEXT_BLOCK)
        self.assertEqual(message, b'')

    def test_union(self):
        blk_owners = ["abc", "def", "wtf"]
        input_hashes = [[h.encode()] for h in blk_owners]
        capnp_impl = CapnpImpl()
        msg_type, message = capnp_impl.get_message(
                               msg_type=MessageType.BLOCK_NOTIFICATION,
                               blockNum=123, blockOwners=blk_owners,
                               inputHashes=input_hashes, newBlock=None)
        self.assertEqual(message.which(), "newBlock")

        wallet = Wallet()
        mtype2, msg2 = capnp_impl.get_signed_message_packed(
                               wallet=wallet,
                               msg_type=MessageType.BLOCK_NOTIFICATION,
                               blockNum=123, blockOwners=blk_owners,
                               inputHashes=input_hashes, newBlock=None)

        mtype3, msg3, sender, timestamp, is_verified = capnp_impl.unpack_message(mtype2, msg2)
        self.assertEqual(is_verified, True)
        self.assertEqual(mtype3, MessageType.BLOCK_NOTIFICATION)
        self.assertEqual(msg3.which(), "newBlock")



from cilantro_ee.services.overlay.new_server import OverlayServer, OverlayClient
from unittest import TestCase
import zmq
import zmq.asyncio
from cilantro_ee.core.crypto.wallet import Wallet
import asyncio
from cilantro_ee.core.sockets.services import _socket, get
from cilantro_ee.core.messages.message import MessageType, Message


async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


class TestOverlayServer(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_serve(self):
        w1 = Wallet()
        o = OverlayServer(
            socket_id=_socket('tcp://127.0.0.1:10999'),
            wallet=w1,
            ctx=self.ctx,
            ip='127.0.0.1',
            peer_service_port=10001,
            event_publisher_port=10002,
            bootnodes=[],
            mn_to_find=[],
            del_to_find=[],
            initial_mn_quorum=0,
            initial_del_quorum=0)

        tasks = asyncio.gather(
            o.serve(),
            stop_server(o, 1)
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

    def test_find_node(self):
        w1 = Wallet()
        o = OverlayServer(
            socket_id=_socket('tcp://127.0.0.1:10999'),
            wallet=w1,
            ctx=self.ctx,
            ip='127.0.0.1',
            peer_service_port=10001,
            event_publisher_port=10002,
            bootnodes=[],
            mn_to_find=[],
            del_to_find=[],
            initial_mn_quorum=0,
            initial_del_quorum=0)

        w2 = Wallet()
        async def lazy_wait():
            await asyncio.sleep(5)

            req = Message.get_signed_message_packed_2(wallet=w2,
                                                      msg_type=MessageType.IP_FOR_VK_REQUEST,
                                                      vk=w1.vk.encode())

            resp = await get(_socket('tcp://127.0.0.1:10999'), msg=req, ctx=self.ctx, dealer=True)
            return resp

        tasks = asyncio.gather(
            o.serve(),
            lazy_wait(),
            stop_server(o, 6)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg_got = res[1]
        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message_2(message=msg_got)

        self.assertEqual(msg.ip.decode(), '127.0.0.1')


class TestOverlayClient(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.ctx.destroy()
        self.loop.close()

    def test_get_ip_for_vk_works(self):
        w1 = Wallet()
        o = OverlayServer(
            socket_id=_socket('tcp://127.0.0.1:10999'),
            wallet=w1,
            ctx=self.ctx,
            ip='127.0.0.1',
            peer_service_port=10001,
            event_publisher_port=10002,
            bootnodes=[],
            mn_to_find=[],
            del_to_find=[],
            initial_mn_quorum=0,
            initial_del_quorum=0)

        w2 = Wallet()
        c = OverlayClient(wallet=w2, ctx=self.ctx, overlay_server_socket=_socket('tcp://127.0.0.1:10999'))

        async def lazy_wait():
            await asyncio.sleep(5)
            return await c.get_ip_for_vk(w1.vk.encode())

        tasks = asyncio.gather(
            o.serve(),
            lazy_wait(),
            stop_server(o, 6)
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        msg = res[1]

        self.assertEqual(msg, '127.0.0.1')



from unittest import TestCase
import unittest
from cilantro_ee.messages.block_data.block_data import GENESIS_BLOCK_HASH
from cilantro_ee.services.storage.state import MetaDataStorage
import json
from cilantro_ee.core.crypto.wallet import Wallet
from cilantro_ee.core.utils.transaction import TransactionBuilder
from cilantro_ee.messages import capnp as schemas
import os
import capnp
import secrets
from cilantro_ee.core.containers.merkle_tree import MerkleTree
from contracting.db import encoder

blockdata_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/blockdata.capnp')
subblock_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/subblock.capnp')
transaction_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/transaction.capnp')
signal_capnp = capnp.load(os.path.dirname(schemas.__file__) + '/signals.capnp')


class TestStateDriver(TestCase):
    def setUp(self):
        self.r = MetaDataStorage()  # this is a class, not an instance, so we do not instantiate
        self.r.flush()

    def tearDown(self):
        self.r.flush()

    def test_state_updated(self):
        # Generate some random transactions
        transactions = []
        get_sets = []
        for i in range(50):
            w = Wallet()
            tx = TransactionBuilder(w.verifying_key(), contract='currency',
                                    function='transfer',
                                    kwargs={'amount': 10, 'to': 'jeff'},
                                    stamps=500000,
                                    processor=secrets.token_bytes(32),
                                    nonce=i)

            tx.sign(w.signing_key())

            tx.proof = b'\x00'
            tx.proof_generated = True

            packed_tx = transaction_capnp.Transaction.from_bytes_packed(tx.serialize())

            # Create a hashmap between a key and the value it should be set to randomly
            get_set = {secrets.token_hex(8): secrets.token_hex(8)}
            get_sets.append(get_set)

            # Put this hashmap as the state of the contract execution and contruct it into a capnp struct
            tx_data = transaction_capnp.TransactionData.new_message(
                transaction=packed_tx,
                status='SUCC',
                state=json.dumps(get_set),
                contractType=0
            )

            # Append it to our list
            transactions.append(tx_data)

        # Build a subblock. One will do
        tree = MerkleTree.from_raw_transactions([tx.to_bytes_packed() for tx in transactions])

        w = Wallet()

        sig = w.sign(tree.root)

        sb = subblock_capnp.SubBlock.new_message(
            merkleRoot=tree.root,
            signatures=[sig],
            merkleLeaves=tree.leaves,
            subBlockIdx=0,
            inputHash=b'a' * 32,
            transactions=[tx for tx in transactions]
        )

        import hashlib

        h = hashlib.sha3_256()
        h.update(b'\00' * 32)
        h.update(tree.root)

        block = blockdata_capnp.BlockData.new_message(
            blockHash=h.digest(),
            blockNum=1,
            blockOwners=[b'\00' * 32],
            prevBlockHash=b'\00' * 32,
            subBlocks=[sb]
        )

        self.r.update_with_block(block)

        for kv in get_sets:
            k, v = list(kv.items())[0]
            got = self.r.get(k)
            self.assertEqual(v, got)

    # TODO test this with publish transactions

    def test_get_latest_block_hash_with_none_set(self):
        b_hash = self.r.get_latest_block_hash()
        self.assertEqual(GENESIS_BLOCK_HASH, b_hash)

    def test_get_latest_block_num_with_none_set(self):
        b_num = self.r.get_latest_block_num()
        self.assertEqual(0, b_num)

    def test_set_get_latest_block_hash(self):
        b_hash = b'A' * 32
        self.r.set_latest_block_hash(b_hash)

        self.assertEqual(self.r.get_latest_block_hash(), b_hash)

    def test_set_get_latest_block_num(self):
        b_num = 9001
        self.r.set_latest_block_num(b_num)

        self.assertEqual(self.r.get_latest_block_num(), b_num)

    def test_set_transaction_data_single_value(self):
        update = {'123': 999}
        encoded = json.dumps(update)
        tx_data = transaction_capnp.TransactionData.new_message(
            state=encoded
        )

        self.r.set_transaction_data(tx=tx_data)

        self.assertEqual(self.r.get('123'), 999)

    def test_set_transaction_multiple_values(self):
        update = {'123': 999,
                  'stu': b'555',
                  'something': [1, 2, 3]}

        print(encoder.encode(update))

        encoded = encoder.encode(update)
        tx_data = transaction_capnp.TransactionData.new_message(
            state=encoded
        )

        self.r.set_transaction_data(tx=tx_data)
        self.assertEqual(self.r.get('123'), 999)
        self.assertEqual(self.r.get('stu'), b'555')
        self.assertEqual(self.r.get('something'), [1, 2, 3])

    def test_set_transaction_corrupted_json_doesnt_work(self):
        update = b'999'
        tx_data = transaction_capnp.TransactionData.new_message(
            state=update
        )
        self.r.set_transaction_data(tx=tx_data)

        # Won't write the data because it's not a valid JSON map
        self.assertEqual(len(self.r.keys()), 0)

    def test_set_transaction_non_dicts_dont_work(self):
        update = ['1', 2, 'three']
        encoded = encoder.encode(update)
        tx_data = transaction_capnp.TransactionData.new_message(
            state=encoded
        )
        self.r.set_transaction_data(tx=tx_data)

        # Won't write the data because it's not a valid JSON map
        self.assertEqual(len(self.r.keys()), 0)




from cilantro_ee.messages.block_data.state_update import FailedBlockNotification
from unittest import TestCase
from unittest.mock import MagicMock, patch


class TestFailedBlockNotification(TestCase):

    @patch("cilantro_ee.messages.block_data.state_update.NUM_SUB_BLOCKS", 4)
    def test_create(self):
        prev_hash = 'A' * 64
        input_hashes = [{'AB' * 32, 'BC' * 32}, {'C'*64, 'D'*64}, set(), {'E'*64}]

        fbn = FailedBlockNotification.create(prev_block_hash=prev_hash, input_hashes=input_hashes, sb_indices=[0, 1])

        self.assertEqual(fbn.prev_block_hash, prev_hash)
        self.assertEqual(fbn.input_hashes, input_hashes)

    @patch("cilantro_ee.messages.block_data.state_update.NUM_SUB_BLOCKS", 4)
    def test_serialize_deserialize(self):
        prev_hash = 'A' * 64
        input_hashes = [{'AB' * 32, 'BC' * 32}, {'C'*64, 'D'*64}, set(), {'E'*64}]

        fbn = FailedBlockNotification.create(prev_block_hash=prev_hash, input_hashes=input_hashes, sb_indices=[0, 1])
        clone = FailedBlockNotification.from_bytes(fbn.serialize())

        self.assertEqual(fbn, clone)



from unittest import TestCase
from cilantro_ee.services.storage.vkbook import VKBook
from contracting.client import ContractingClient


class TestVKBook(TestCase):
    def setUp(self):
        self.client = ContractingClient()
        self.client.flush()

    def tearDown(self):
        self.client.flush()

    def test_submit_new_vkbook(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.masternodes, masternodes)
        self.assertEqual(v.delegates, delegates)
        self.assertEqual(v.stamps_enabled, stamps)
        self.assertEqual(v.nonces_enabled, nonces)

    def test_reinitialization_does_not_resubmit_and_overwrite(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        new_masternodes = ['d', 'e', 'f']
        new_delegates = ['a', 'b', 'c']
        new_stamps = True
        new_nonces = True

        v = VKBook(new_masternodes, new_delegates, stamps=new_stamps, nonces=new_nonces, debug=False)

        self.assertEqual(v.masternodes, masternodes)
        self.assertEqual(v.delegates, delegates)
        self.assertEqual(v.stamps_enabled, stamps)
        self.assertEqual(v.nonces_enabled, nonces)

    def test_witnesses_as_default(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.witnesses, [])

    def test_notifiers_as_default(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.notifiers, [])

    def test_schedulers_as_default(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.schedulers, [])

    def test_state_sync_returns_masternodes_and_delegates(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.state_sync, masternodes + delegates)

    def test_all_returns_masternodes_delegates_and_witnesses(self):
        masternodes = ['a', 'b', 'c']
        delegates = ['d', 'e', 'f']
        stamps = False
        nonces = False

        v = VKBook(masternodes, delegates, stamps=stamps, nonces=nonces, debug=False)

        self.assertEqual(v.all, masternodes + delegates + [])



"""
    SubBlockBuilder

    If Block is viewed as consists of a merkle tree of transactions, then sub-block refers to the sub-tree of the block.
    Conceptually Sub Block could form whole block or part of block. This lets us scale things horizontally.
    Each of this SB builder will be started on a separate process and will coordinate with BlockManager
    to resolve db conflicts between sub-blocks and send resolved sub-block to master.
    It also sends in partial data of transactions along with the sub-block

    We typically take all transactions from a single master to form a sub-block,
    but a sub-block builder can be responsible for more than one master and so can make more than one sub-block.
    This ensures our ordering guarantees that transactions entered at a master is executed in that order,
    but we will decide the order of transactions between different masters.

"""


# need to clean this up - this is a dirty version of trying to separate out a sub-block builder in the old code

from typing import Dict, Callable
from cilantro_ee.services.storage.state import MetaDataStorage
from cilantro_ee.constants.zmq_filters import *
from cilantro_ee.constants.system_config import *

from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core.messages.message import Message
from contracting.config import NUM_CACHES
from contracting.stdlib.bridge.time import Datetime
from contracting.db.cr.client import SubBlockClient
from contracting.db.cr.callback_data import SBData

from cilantro_ee.core.utils.worker import Worker
from cilantro_ee.core.utils.network_topology import NetworkTopology

from cilantro_ee.core.containers.merkle_tree import MerkleTree
from cilantro_ee.core.containers.linked_hashtable import LinkedHashTable

from cilantro_ee.core.utils.transaction import transaction_is_valid, TransactionException

from cilantro_ee.utils.hasher import Hasher
from cilantro_ee.core.crypto.wallet import _verify
from enum import Enum, unique
import asyncio, zmq.asyncio, time
from datetime import datetime
import hashlib
from decimal import Decimal

from cilantro_ee.core.nonces import NonceManager

class Payload:
    def __init__(self, sender, nonce, processor, stamps_supplied, contract_name, function_name, kwargs):
        self.sender = sender
        self.nonce = nonce
        self.processor = processor
        self.stampsSupplied = stamps_supplied
        self.contractName = contract_name
        self.functionName = function_name
        self.kwargs = kwargs


class UnpackedContractTransaction:
    def __init__(self, capnp_struct):
        kwargs = {}
        for entry in capnp_struct.payload.kwargs.entries:
            if entry.value.which() == 'fixedPoint':
                kwargs[entry.key] = Decimal(entry.value.fixedPoint)
            else:
                kwargs[entry.key] = getattr(entry.value, entry.value.which())

        self.payload = Payload(sender=capnp_struct.payload.sender,
                               nonce=capnp_struct.payload.nonce,
                               processor=capnp_struct.payload.processor,
                               stamps_supplied=capnp_struct.payload.stampsSupplied,
                               contract_name=capnp_struct.payload.contractName,
                               function_name=capnp_struct.payload.functionName,
                               kwargs=kwargs)

@unique
class NextBlockState(Enum):
    NOT_READY = 0
    READY = 1
    PROCESSED = 2


class SBClientManager:
    def __init__(self, sbb_idx, loop):
        # self.client = SubBlockClient(sbb_idx=sbb_idx, num_sbb=NUM_SB_PER_BLOCK, loop=loop)
        self.next_sm_index = 0
        self.max_caches = 2
        self.sb_caches = []


class NextBlockToMake:
    def __init__(self, block_index: int=0, state: NextBlockState=NextBlockState.READY):
        self.pending_sm_idx = 0
        self.to_finalize_txs = LinkedHashTable()
        self.next_block_index = block_index
        self.state = state


# This is a convenience struct to hold all data related to a sub-block in one place.
# Since we have more than one sub-block per process, SBB'er will hold an array of SBManager objects
class SBManager:
    def __init__(self, sub_block_index: int, sub_socket, processed_txs_timestamp: int=0):
        self.sub_block_index = sub_block_index
        self.connected_vk = None
        self.empty_input_iter = 0
        self.sub_socket = sub_socket
        self.processed_txs_timestamp = processed_txs_timestamp
        self.pending_txs = LinkedHashTable()

    def get_empty_input_hash(self):
        self.empty_input_iter += 1
        return Hasher.hash(self.connected_vk + str(self.empty_input_iter), return_bytes=True)


class SubBlockBuilder(Worker):
    def __init__(self, ip: str, signing_key: str, ipc_ip: str, ipc_port: int, sbb_index: int, *args, **kwargs):
        super().__init__(signing_key=signing_key, name="SubBlockBuilder_{}".format(sbb_index))

        self.state = MetaDataStorage()

        self.ip = ip
        self.sb_blder_idx = sbb_index
        self.num_txn_bags = 0
        self._empty_txn_batch = []

        self.client = SubBlockClient(sbb_idx=sbb_index, num_sbb=NUM_SB_PER_BLOCK, loop=self.loop)
        self.nonce_manager = NonceManager()

        # DEBUG -- TODO DELETE
        self.log.important("num sbb per blk {}".format(NUM_SB_PER_BLOCK))
        # END DEBUG

        # Create DEALER socket to talk to the BlockManager process over IPC
        self.ipc_dealer = self.manager.create_socket(socket_type=zmq.DEALER,
                                                     name="SBB-IPC-Dealer[{}]".format(self.sb_blder_idx), secure=False)
        # setting identity and other (few important) socket options can be part of create_socket api above
        self.ipc_dealer.setsockopt(zmq.IDENTITY, str(self.sb_blder_idx).encode())
        # connect or bind is a separate step
        self.ipc_dealer.connect(port=ipc_port, protocol='ipc', ip=ipc_ip)

        self.tasks.append(self.ipc_dealer.add_handler(handler_func=self.handle_ipc_msg))
        # Adding message_handler with dictionary of actions as a separate step - this api could change
        # self.tasks.append(self.ipc_dealer.add_message_handler(self.get_ipc_message_action_dict()))

        # BIND sub sockets to listen to witnesses
        self.sb_managers = []
        self._create_sub_sockets()
        # need to tie with catchup state to initialize to real next_block_to_make
        self._next_block_to_make = NextBlockToMake()
        self.tasks.append(self._connect_and_process())

        self.log.notice("sbb_index {} tot_sbs {} num_blks {} num_sb_blders {} num_sb_per_block {} num_sb_per_builder {} sbs_per_blk_per_blder {}"
                        .format(sbb_index, NUM_SUB_BLOCKS, NUM_BLOCKS, NUM_SB_BUILDERS, NUM_SB_PER_BLOCK, NUM_SB_PER_BUILDER, NUM_SB_PER_BLOCK_PER_BUILDER))

        # raghu todo - this added the requirement that sb per block per builder has to be 1
        self.pending_transactions = []

        self.run()

    def run(self):
        self.log.notice("SBB {} starting...".format(self.sb_blder_idx))
        self.loop.run_until_complete(asyncio.gather(*self.tasks))

    def get_ipc_message_action_dict(self):
        return {
            MessageType.MAKE_NEXT_BLOCK: self.__make_next_sub_block,
            MessageType.BLOCK_NOTIFICATION: self.__fail_block
        }

    async def _connect_and_process(self):
        # first make sure, we have overlay server ready
        await self._wait_until_ready()
        await self._connect_sub_sockets()
        mtype, msg = Message.get_message_packed(MessageType.READY)
        await self.ipc_dealer.send_multipart([mtype, msg])

    # raghu todo - call this right after catch up phase, need to figure out the right input hashes though for next block
    def move_next_block_to_make(self):
        if self._next_block_to_make.state == NextBlockState.PROCESSED:
            self._next_block_to_make.next_block_index = self._next_block_to_make.next_block_index + 1 % NUM_BLOCKS
            self._next_block_to_make.state = NextBlockState.READY

        return self._next_block_to_make.state == NextBlockState.READY

    def move_pending_smi(self):
            smi = self._next_block_to_make.pending_sm_idx + NUM_SB_PER_BLOCK_PER_BUILDER
            self._next_block_to_make.pending_sm_idx = smi % len(self.sb_managers)

    def reset_next_block_to_make(self):
            self.move_pending_smi()
            self._next_block_to_make.next_block_index = self._next_block_to_make.pending_sm_idx // NUM_SB_PER_BLOCK_PER_BUILDER
            self._next_block_to_make.state = NextBlockState.READY
            

    def _create_sub_sockets(self):
        for idx in range(NUM_SB_PER_BUILDER):
            sub = self.manager.create_socket(socket_type=zmq.SUB, name="SBB-Sub[{}]-{}".format(self.sb_blder_idx, idx),
                                             secure=True)

            sub.setsockopt(zmq.SUBSCRIBE, TRANSACTION_FILTER.encode())
            sb_index = idx * NUM_SB_BUILDERS + self.sb_blder_idx

            self.sb_managers.append(SBManager(sub_block_index=sb_index, sub_socket=sub))
            self.tasks.append(sub.add_handler(handler_func=self.handle_sub_msg, handler_key=idx))

    async def _connect_sub_sockets(self):
        for smi, d in enumerate(NetworkTopology.get_sbb_publishers(self.verifying_key, self.sb_blder_idx), 0):
            vk, port, sb_idx = d['vk'], d['port'], d['sb_idx']
            assert self.sb_managers[smi].sub_block_index == sb_idx, "something went wrong in connections"
            self.sb_managers[smi].sub_socket.connect(port=port, vk=vk)
            self.sb_managers[smi].connected_vk = vk

    def _align_to_hash(self, smi, input_hash):
        num_discards = 0

        if input_hash in self.sb_managers[smi].pending_txs:
            # clear entirely to_finalize
            num_discards = num_discards + len(self.sb_managers[smi].to_finalize_txs)
            self.sb_managers[smi].to_finalize_txs.clear()
            ih2 = None

            while input_hash != ih2:
                ih2, txs_bag = self.sb_managers[smi].pending_txs.pop_front()

                self.adjust_work_load(txs_bag, False)

                num_discards = num_discards + 1

        elif input_hash in self.sb_managers[smi].to_finalize_txs:
            ih2 = None
            while input_hash != ih2:
                ih2, txs_bag = self.sb_managers[smi].to_finalize_txs.pop_front()
                num_discards = num_discards + 1
        return num_discards

    def _align_to_hashes(self, smi, input_hashes):
        ih, ib = self._next_block_to_make.to_finalize_txs.pop_front()
        if ih in input_hashes:
            return
        for ihash in input_hashes:
            if ihash in self._next_block_to_make.to_finalize_txs:
                self.log.error("finalize transaction queue is inconsistent!")
                while ihash in self._next_block_to_make.to_finalize_txs:
                    self._next_block_to_make.to_finalize_txs.pop_front()
                return
            if ihash in self.sb_managers[smi].pending_txs:
                while ihash in self.sb_managers[smi].pending_txs:
                    self.sb_managers[smi].pending_txs.pop_front()
                return
        # self.sb_managers[smi].pending_txs.insert_front(ih, ib)

    def _return_excess_queue(self, smi: int):
        num_txns = len(self._next_block_to_make.to_finalize_txs)
        num_blocks = num_txns // NUM_SB_PER_BLOCK_PER_BUILDER
        if num_blocks < 1 or num_txns != num_blocks * NUM_SB_PER_BLOCK_PER_BUILDER:
            self.log.error("finalize txn queue is inconsistent!")
            return
        if num_blocks == 1:
            return
        last_smi = smi + num_blocks * NUM_SB_PER_BLOCK_PER_BUILDER - 1
        ret_smi = last_smi % len(self.sb_managers)
        while num_txns >= NUM_SB_PER_BLOCK_PER_BUILDER:
            ih, ib = self._next_block_to_make.to_finalize_txs.pop_back()
            self.sb_managers[ret_smi].pending_txs.insert_front(ih, ib)
            num_txns -= 1
            ret_smi -= 1
            
    def _fail_block(self, block):
        self.log.notice("FailedBlockNotification - aligning input hashes")
        self.client.flush_all()

        # Toss all pending nonces
        self.nonce_manager.delete_pending_nonces()

        smi = block.firstSbIdx // NUM_SB_BUILDERS
        # assert smi == self._next_block_to_make.pending_sm_idx, "misalignment in align input hashes"
        if smi != self._next_block_to_make.pending_sm_idx:
            self.log.warning("misalignment in align input hashes - smi {} "
                  "pending smi {}".format(smi, self._next_block_to_make.pending_sm_idx))
            self._next_block_to_make.pending_sm_idx = smi

        self._return_excess_queue(smi)

        # first return more than one block worth of sub-blocks to the pending txs
        # then align
        #  if input hash matches one in pending txs, then pop it from pending txs and delete the top from to_finalize
        #  else if input hash matches the top one in to_finalize, then just pop it
        #  else if input hash matches one that is not the top one, then log the error and pop it to that point
        #  else if input hash is not matching anything, then return the top one to pending-txs (to redo same sub-block)
        
        for i in range(NUM_SB_PER_BLOCK_PER_BUILDER):
            idx = self.sb_blder_idx + i * NUM_SB_BUILDERS
            input_hashes = block.inputHashes[idx]
            self._align_to_hashes(smi + i, input_hashes)

        self.reset_next_block_to_make()

        # self._make_next_sb()

    async def __fail_block(self, msg: bytes):
        block = BlockNotification.unpack_block_notification(msg)
        self._fail_block(block)

    def handle_ipc_msg(self, frames):
        self.log.info("SBB received an IPC message {}".format(frames))
        assert len(frames) == 2, "Expected 2 frames: (msg_type, msg_blob). Got {} instead.".format(frames)

        msg_type = frames[0]
        msg_blob = frames[1]

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        self.log.info("Got message type '{}' msg '{}' from block manager. "
                      .format(msg_type, msg))
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        if msg_type == MessageType.MAKE_NEXT_SB:
            self._make_next_sub_block()
        elif msg_type == MessageType.COMMIT_CUR_SB:
            self._commit_cur_sb()
        # elif msg_type == MessageType.DISCORD_AND_ALIGN:
        elif msg_type == MessageType.BLOCK_NOTIFICATION:
            self._fail_block(msg)
            return

        else:
            self.log.error("Got invalid message type '{}' from block manager. "
                           "Ignoring it ..".format(type(msg_type)))

    def adjust_work_load(self, txn_batch, is_add: bool):
        if len(txn_batch) == 0:
            return

        self.num_txn_bags += 1 if is_add else -1

        assert self.num_txn_bags >= 0, "Something went wrong!"

        # Send pending work signal
        if self.num_txn_bags == 0:
            msg_type, msg = Message.get_message_packed(MessageType.NO_TRANSACTIONS)
            self.ipc_dealer.send_multipart([msg_type, msg])

        elif is_add and self.num_txn_bags == 1:
            msg_type, msg = Message.get_message_packed(MessageType.PENDING_TRANSACTIONS)
            self.ipc_dealer.send_multipart([msg_type, msg])

    # ONLY FOR TX BATCHES
    def handle_sub_msg(self, frames, index):
        msg_filter, msg_type, msg_blob = frames

        msg_type, msg, sender, timestamp, is_verified = Message.unpack_message(msg_type, msg_blob)
        if not is_verified:
            self.log.error("Failed to verify the message of type {} from {} at {}. Ignoring it .."
                          .format(msg_type, sender, timestamp))
            return

        if msg_type == MessageType.TRANSACTION_BATCH and 0 <= index < len(self.sb_managers):

            batch = msg
            timestamp = batch.timestamp


            self.log.info('Got tx batch with {} txs with input hash {} ts {} for sbb {}'
                          .format(len(batch.transactions), batch.inputHash.hex(), timestamp, index))

            if batch.sender.hex() not in PhoneBook.masternodes:
                self.log.critical('RECEIVED TX BATCH FROM NON MASTER NODE')
                return

            else:
                self.log.success('{} is a masternode!'.format(batch.sender.hex()))

            if timestamp <= self.sb_managers[index].processed_txs_timestamp:
                self.log.debug("Got timestamp {} that is prior to the most recent timestamp {} for sb_manager {} tho"
                               .format(timestamp, self.sb_managers[index].processed_txs_timestamp, index))
                return

            # Set up a hasher for input hash and a list for valid txs
            h = hashlib.sha3_256()
            valid_transactions = []

            for tx in batch.transactions:
                # Double check to make sure all transactions are valid
                try:
                    transaction_is_valid(tx=tx,
                                         expected_processor=batch.sender,
                                         driver=self.nonce_manager,
                                         strict=False)
                    valid_transactions.append(tx)
                except TransactionException:
                    self.log.critical('TX NOT VALID FOR SOME REASON.')

                # Hash all transactions regardless because the proof from masternodes is derived from all hashes
                h.update(tx.as_builder().to_bytes_packed())

            h.update('{}'.format(batch.timestamp).encode())
            input_hash = h.digest()
            if input_hash != batch.inputHash or \
               not _verify(batch.sender, h.digest(), batch.signature):
                self.log.critical("Transaction batch and its input hash {} with"
                          " timestamp {} doesn't match the signature from sender"
                          " {}".format(input_hash, batch.timestamp, batch.sender))
                return

            # if the sb_manager already has this bag, ignore it
            if input_hash in self.sb_managers[index].pending_txs:
                self.log.debugv("Input hash {} already found in sb_manager at index {}".format(input_hash, index))
                return

            # DEBUG -- TODO DELETE
            self.log.notice("Recv tx batch w/ {} transactions, and input hash {}".format(len(batch.transactions), input_hash))

            self.sb_managers[index].processed_txs_timestamp = timestamp

            self.log.info("Queueing transaction batch for sb manager {}. SB_Manager={}".format(index, self.sb_managers[index]))
            self.adjust_work_load(valid_transactions, True)

            # Add the valid transactions to
            self.sb_managers[index].pending_txs.append(input_hash, valid_transactions)

    def _create_empty_sbc(self, sb_idx: int, sb_data: SBData):
        """
        Creates an Empty Sub Block Contender
        """
        self.log.info("Building empty sub block contender for input hash {}".format(sb_data.input_hash))

        self.log.info(type(sb_data.input_hash))

        if type(sb_data.input_hash) == str:
            input_hash = bytes.fromhex(sb_data.input_hash)
        else:
            input_hash = sb_data.input_hash

        _, merkle_proof = Message.get_message_packed(
                                    MessageType.MERKLE_PROOF,
                                    hash=input_hash,
                                    signer=self.wallet.verifying_key(),
                                    signature=self.wallet.sign(input_hash))

        mtype, msg = Message.get_message_packed(MessageType.SUBBLOCK_CONTENDER,
                                resultHash=input_hash,
                                inputHash=input_hash,
                                merkleLeaves=[],
                                signature=merkle_proof,
                                transactions=[],
                                subBlockIdx=sb_idx,
                                prevBlockHash=self.state.get_latest_block_hash())

        self.log.important2("Sending EMPTY SBC with input hash {} to block manager!".format(input_hash))

        self.ipc_dealer.send_multipart([mtype, msg])

    def _create_sbc_from_batch(self, sb_idx: int, sb_data: SBData):

        """
        Creates a Sub Block Contender from a TransactionBatch
        """

        self.log.info("Building sub block contender for input hash {}".format(sb_data.input_hash))
        exec_data = sb_data.tx_data

        txs_data = []

        # Add stamps used to TransactionData payload
        for i in range(len(exec_data)):
            d = exec_data[i]
            tx = self.pending_transactions[i]
            _, txn_msg = Message.get_message_packed(
                                    MessageType.TRANSACTION_DATA,
                                    transaction=tx,
                                    status=d.status,
                                    state=d.state)
            txs_data.append(txn_msg)

        # build sbc
        merkle = MerkleTree.from_raw_transactions(txs_data)

        _, merkle_proof = Message.get_message_packed(
                                    MessageType.MERKLE_PROOF,
                                    hash=merkle.root,
                                    signer=self.wallet.verifying_key(),
                                    signature=self.wallet.sign(merkle.root))

        mtype, sbc = Message.get_message_packed(
                           MessageType.SUBBLOCK_CONTENDER,
                           resultHash=merkle.root,
                           inputHash=sb_data.input_hash,
                           merkleLeaves=[leaf for leaf in merkle.leaves],
                           signature=merkle_proof,
                           transactions=[tx for tx in txs_data],
                           subBlockIdx=sb_idx,
                           prevBlockHash=self.state.get_latest_block_hash())

        self.pending_transactions = []
        self.ipc_dealer.send_multipart([mtype, sbc])


    def create_sb_contender(self, sb_idx: int, sb_data: SBData):
        if len(sb_data.tx_data) > 0:
            self._create_sbc_from_batch(sb_idx, sb_data)
        else:
            self._create_empty_sbc(sb_idx, sb_data)


    # raghu todo sb_index is not correct between sb-builder and seneca-client. Need to handle more than one sb per client?
    def _execute_sb(self,
                    input_hash: bytes,
                    tx_batch: list,
                    timestamp: float,
                    sb_idx: int):

        self.log.debug("SBB {} attempting to build a sub block with index {}"
                       .format(self.sb_blder_idx, sb_idx))

        callback = self.create_sb_contender

        # Pass protocol level variables into environment so they are accessible at runtime in smart contracts
        block_hash = self.state.latest_block_hash
        block_num = self.state.latest_block_num

        dt = datetime.utcfromtimestamp(timestamp)
        dt_object = Datetime(year=dt.year,
                             month=dt.month,
                             day=dt.day,
                             hour=dt.hour,
                             minute=dt.minute,
                             second=dt.second,
                             microsecond=dt.microsecond)

        environment = {
            'block_hash': block_hash,
            'block_num': block_num,
            'now': dt_object
        }

        self.log.info('Transactions to execute: {}'.format(tx_batch))

        transactions = []
        for transaction in tx_batch:
            # This should be streamlined so that we can just pass the tx_batch forward because it's already ready to be processed at this point
            # The reason why it isn't like this already is because Contracting uses a weird pseudo wrapper for the capnp struct
            transactions.append(UnpackedContractTransaction(transaction))
            self.pending_transactions.append(transaction)

        result = self.client.execute_sb(input_hash,
                                        transactions,
                                        sb_idx,
                                        callback,
                                        environment=environment)

        self.log.success('RESULT FOR TX BATCH: {}'.format(result))

        if result:
            self._next_block_to_make.state = NextBlockState.PROCESSED
            return True

        return False

    def _make_next_sb(self):
        self.log.info('making next sb')
        if not self.move_next_block_to_make():
            self.log.info("Not ready to make next sub-block. Waiting for seneca-client to be ready ... ")
            return

        # now start next one
        cur_block_index = self._next_block_to_make.next_block_index
        self.log.info('Working on {}'.format(cur_block_index))

        sm_idx_start = cur_block_index * NUM_SB_PER_BLOCK_PER_BUILDER

        for i in range(NUM_SB_PER_BLOCK_PER_BUILDER):
            sm_idx = sm_idx_start + i

            if sm_idx >= len(self.sb_managers):    # out of range already
                self.log.warning("Uneven sub-blocks per block. May not work in "
                           "seneca clients properly in current scheme. sm idx "
                           "{} num_sb_pb_pb {} num_sb_mgrs {}".format(sm_idx, \
                           NUM_SB_PER_BLOCK_PER_BUILDER, len(self.sb_managers)))
                return

            sb_index = self.sb_managers[sm_idx].sub_block_index
            timestamp = self.sb_managers[sm_idx].processed_txs_timestamp

            if len(self.sb_managers[sm_idx].pending_txs) > 0:
                input_hash, tx_batch = self.sb_managers[sm_idx].pending_txs.pop_front()
            else:
                timestamp = float(time.time())
                input_hash = self.sb_managers[sm_idx].get_empty_input_hash()
                tx_batch = self._empty_txn_batch

            self.log.info("Make next sub-block with input hash {}".format(input_hash.hex()))
            self._execute_sb(input_hash, tx_batch, timestamp, sb_index)
            self._next_block_to_make.to_finalize_txs.append(input_hash, tx_batch)

    def _update_master_db(self):
        self.client.update_master_db()
        for _ in range(NUM_SB_PER_BLOCK_PER_BUILDER):
            input_hash, tx_batch = self._next_block_to_make.to_finalize_txs.pop_front()
            self.adjust_work_load(tx_batch, False)
        self.move_pending_smi()

    def _commit_cur_sb(self):
        self.log.info("Merge pending db to master db")
        self._update_master_db()

    def _make_next_sub_block(self):
        self._make_next_sb()

    async def __make_next_sub_block(self, msg: bytes):
        self._make_next_sub_block()



from unittest import TestCase
from cilantro_ee.core.crypto.wallet import Wallet


class TestWallet(TestCase):
    def test_init_new_wallets(self):
        Wallet()

    def test_init_wallet_with_seed_returns_deterministic_wallet(self):
        w = Wallet()

        a = Wallet(seed=w.signing_key())

        self.assertEqual(w.vk, a.vk)
        self.assertEqual(w.sk, a.sk)
        self.assertEqual(w.zmq_key, a.zmq_key)

    def test_signing_key_as_bytes(self):
        w = Wallet()

        _w = w.signing_key()

        self.assertTrue(isinstance(_w, bytes))

        _h = w.signing_key(as_hex=True)

        self.assertTrue(isinstance(_h, str))

    def test_verifying_key_as_bytes(self):
        w = Wallet()

        _w = w.verifying_key()

        self.assertTrue(isinstance(_w, bytes))

        _h = w.verifying_key(as_hex=True)

        self.assertTrue(isinstance(_h, str))

    def test_sign_string_breaks(self):
        w = Wallet()

        with self.assertRaises(AssertionError):
            w.sign('hello')

    def test_sign_bytes_returns_signature(self):
        w = Wallet()

        signature = w.sign(b'hello')

        self.assertTrue(isinstance(signature, bytes))
        self.assertEqual(len(signature), 64)

    def test_sign_bytes_returns_hex_signature(self):
        w = Wallet()

        signature = w.sign(b'hello', as_hex=True)

        self.assertTrue(isinstance(signature, str))
        self.assertEqual(len(signature), 128)

    def test_signature_with_correct_message_returns_true(self):
        w = Wallet()

        message = b'howdy'
        signature = w.sign(message)

        self.assertTrue(w.verify(message, signature))

    def test_signature_with_wrong_message_returns_false(self):
        w = Wallet()

        message = b'howdy'
        signature = w.sign(message)

        self.assertFalse(w.verify(b'hello', signature))

    def test_signature_with_wrong_vk_returns_false(self):
        w = Wallet()

        message = b'howdy'
        signature = w.sign(message)

        a = Wallet()

        self.assertFalse(a.verify(message, signature))



from vmnet.testcase import BaseNetworkTestCase
import unittest, time, random, vmnet, cilantro_ee
from os.path import join, dirname
from cilantro_ee.utils.test.mp_test_case import vmnet_test


LOG_LEVEL = 0

def wrap_func(fn, *args, **kwargs):
    def wrapper():
        return fn(*args, **kwargs)
    return wrapper

def run_mn(slot_num):
    from cilantro_ee.core.logger import get_logger, overwrite_logger_level
    from cilantro_ee.nodes import NodeFactory
    from cilantro_ee.constants.testnet import TESTNET_MASTERNODES

    import os
    import logging

    # overwrite_logger_level(logging.WARNING)
    overwrite_logger_level(15)

    mn_info = TESTNET_MASTERNODES[slot_num]
    mn_info['ip'] = os.getenv('HOST_IP')

    NodeFactory.run_masternode(signing_key=mn_info['sk'], ip=mn_info['ip'], reset_db=True)


def trigger_mn_wr(blocks):
    from cilantro_ee.core.logger import get_logger, overwrite_logger_level
    from cilantro_ee.utils.test import God
    overwrite_logger_level(15)
    God.dump_it(volume=volume, delay=delay)


class MasterStore(BaseNetworkTestCase):

    BLOCKS = 1000
    config_file = join(dirname(cilantro_ee.__path__[0]), 'vmnet_configs', 'cilantro_ee-mn.json')

    @vmnet_test(run_webui=True)
    def mn_store(self):

        # Bootstrap TESTNET_MASTERNODES
        self.execute_python('masternode', run_mn, profiling=None)

        self.execute_python('mgmt', wrap_func(trigger_mn_wr, blocks=self.BLOCKS))

        input("Enter any key to terminate")

if __name__ == '__main__':
    unittest.main()



value = Variable()
current_votes = Hash(default_value=0)
has_voted = Hash()

last_election = Variable()
election_start = Variable()

election_length = datetime.DAYS * 1
election_interval = datetime.WEEKS * 1

@construct
def seed():
    value.set([0.5, 0.5, 0, 0])
    last_election.set(now)
    election_start.set(None)

@export
def current_value():
    return value.get()

@export
def vote(vk, obj):
    if election_start.get() is not None:
        tally_vote(vk, obj)

        # If it has been over a day since the election started... End the election
        if now - election_start.get() >= election_length:
            # Calculate ratio of votes
            masternode_votes = current_votes['masternodes'] or 1
            delegate_votes = current_votes['delegates'] or 1
            blackhole_votes = current_votes['blackhole'] or 1
            foundation_votes = current_votes['foundation'] or 1

            total_votes = masternode_votes + delegate_votes + blackhole_votes + foundation_votes

            # Do the same for each party before dividing
            mn = masternode_votes / total_votes
            dl = delegate_votes / total_votes
            bh = blackhole_votes / total_votes
            fd = foundation_votes / total_votes

            # Set the new value
            value.set([mn, dl, bh, fd])

            # Reset everything
            election_start.set(None)
            last_election.set(now)
            current_votes.clear()
            has_voted.clear()

    # If its been 1 week since the last election ended... Start the election
    elif now - last_election.get() > election_interval:
        # Set start to now
        election_start.set(now)
        current_votes.clear()
        tally_vote(vk, obj)


def tally_vote(vk, obj):
    assert vote_is_valid(obj), 'Invalid vote object passed!'
    assert has_voted[vk] is None, 'VK has already voted!'

    has_voted[vk] = True

    a, b, c, d = obj

    current_votes['masternodes'] += a
    current_votes['delegates'] += b
    current_votes['blackhole'] += c
    current_votes['foundation'] += d


def vote_is_valid(obj):
    if type(obj) != list:
        return False

    if len(obj) != 4:
        return False

    s = 0
    for o in obj:
        if type(o) != int:
            return False
        if o < 0:
            return False
        s += o

    if s != 100:
        return False

    return True



import zmq, zmq.asyncio, asyncio, traceback, time
from os import getenv as env
from cilantro_ee.constants.overlay_network import *
from cilantro_ee.constants.ports import AUTH_PORT
from zmq.auth.thread import ThreadAuthenticator
from zmq.auth.asyncio import AsyncioAuthenticator
# from cilantro_ee.services.overlay.ip import *
# from cilantro_ee.core.sockets.socket_auth import SocketAuth
from cilantro_ee.core.sockets.socket import SocketUtil
from cilantro_ee.utils.keys import Keys
from cilantro_ee.core.logger import get_logger
from cilantro_ee.services.storage.vkbook import VKBook, PhoneBook
from collections import defaultdict
from cilantro_ee.constants import conf


class Handshake:
    def __init__(self, vk, ctx):
        self.vk = vk
        self.ctx = ctx

        self.log = get_logger('OS.Handshake')
        self.host_ip = conf.HOST_IP
        self.port = AUTH_PORT
        self.url = 'tcp://*:{}'.format(self.port)
        self.identity = '{}:{}'.format(self.vk, self.host_ip)
        
        self.server_sock = SocketUtil.create_socket(self.ctx, zmq.ROUTER)
        # not setting the identity for this socket as communication to it is always 1 on 1
        self.server_sock.curve_secretkey = Keys.private_key
        self.server_sock.curve_publickey = Keys.public_key
        self.server_sock.curve_server = True
        self.server_sock.bind(self.url)

    async def authenticate(self, event_id, vk, ip, domain, is_first_time):
        assert ip != self.host_ip, "Silly to authenticate with yourself! Check your logic!"

        start = time.time()
        authorized = False
        url = 'tcp://{}:{}'.format(ip, self.port)
        self.log.info('Sending handshake request from {} to {} (vk={})'.format(self.host_ip, ip, vk))
        client_sock = SocketUtil.create_socket(self.ctx, zmq.DEALER)  # could be simple req socket too
        client_sock.setsockopt(zmq.IDENTITY, '{}:{}'.format(self.identity, event_id).encode())
        # raghu TODO do we really need when it is bypassing the check anyway
        client_sock.curve_secretkey = Keys.private_key
        client_sock.curve_publickey = Keys.public_key
        client_sock.curve_serverkey = Keys.vk2pk(vk)
        client_sock.connect(url)
        client_sock.send_multipart([domain.encode()])

        try:
            domain = [chunk.decode() for chunk in await asyncio.wait_for(client_sock.recv_multipart(), AUTH_TIMEOUT)][0]
            self.log.info('Received a handshake reply from {} to {} (vk={})'.format(ip, self.host_ip, vk))
            authorized = self.process_handshake(ip, vk, domain)
            self.log.notice('Complete (took {}s): {} <=o= {} (vk={})'.format(time.time()-start, self.host_ip, ip, vk))
        except asyncio.TimeoutError:
            self.log.warning('Timeout (took {}s): {} <=:= {} (vk={})'.format(time.time()-start, self.host_ip, ip, vk))
        except Exception:
            self.log.error(traceback.format_exc())
        finally:
            client_sock.close()

        return authorized

    async def listen(self):
        self.log.info('Listening to other nodes on {}'.format(self.url))
        while True:
            try:
                vk_ip, domain = [chunk.decode() for chunk in await self.server_sock.recv_multipart()]
                # raghu todo also recv is_first_time flag and announce it via evt_socket
                vk, ip, event_id = vk_ip.split(':')
                self.log.info('Received a handshake request from {} (vk={}) with event_id {}'.format(ip, vk, event_id))
                authorized = self.process_handshake(ip, vk, domain)
                if authorized:
                    self.server_sock.send_multipart([vk_ip.encode(), domain.encode()])
            except Exception as e:
                self.log.error(traceback.format_exc())

    def process_handshake(self, ip, vk, domain):
        if self.validate_roles_with_domain(domain, vk):
            self.log.info('Authorized: {} <=O= {} (vk={}, domain={})'.format(self.host_ip, ip, vk, domain))
            return True
        else:
            # NOTE The sender proved that it has the VK via the router's identity frame but the sender is not found in the receiver's VKBook
            # raghu todo - could be a potential audit layer log where this unauthorized ip is blocked at some point ?
            self.log.warning('Unauthorized VK: {} <=X= {} (vk={}, domain={})'.format(self.host_ip, ip, vk, domain))
            return False

    @staticmethod
    def validate_roles_with_domain(domain, vk, roles='any'):
        if roles == 'any':
            return vk in PhoneBook.all
        else:
            if 'masternodes' in roles and vk in PhoneBook.masternodes:
                return True
            if 'witnesses' in roles and vk in PhoneBook.witnesses:
                return True
            if 'delegates' in roles and vk in PhoneBook.delegates:
                return True
        return False



from cilantro_ee.core.logger.base import get_logger
from cilantro_ee.utils.test.dumpatron import Dumpatron
from cilantro_ee.utils.test.god import God
from cilantro_ee.utils.test.wallets import GENERAL_WALLETS
import time, random, asyncio, aiohttp
from collections import defaultdict


class CurrencyTester(Dumpatron):

    FETCH_BALANCES_TIMEOUT = 30
    ASSERT_BALANCES_TIMEOUT = 90
    SLEEP_BEFORE_ASSERTING = 12  # How long we should wait between sending transactions and asserting _new balances
    POLL_INTERVAL = 2

    MIN_AMOUNT = 1
    MAX_AMOUNT = 1000000

    def __init__(self, *args, wallets=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.log = get_logger("CurrencyTester")

        # For keeping track of wallet balances and asserting the correct amnt was deducted
        self.wallets = wallets or GENERAL_WALLETS
        self.all_vks = set([w[1] for w in self.wallets])
        self.deltas = defaultdict(int)
        self.init_balances = {}

    async def _start(self):
        await self.get_initial_balances()
        await self.send_test_currency_txs()
        self.log.test("Sleeping for {} seconds before checking assertions...".format(self.SLEEP_BEFORE_ASSERTING))
        await asyncio.sleep(self.SLEEP_BEFORE_ASSERTING)
        await self.assert_balances_updated()

    async def get_initial_balances(self):
        self.log.test("Getting initial balances for {} wallets...".format(len(self.wallets)))
        elapsed = 0
        while len(self.wallets) != len(self.init_balances):
            await asyncio.sleep(self.POLL_INTERVAL)
            elapsed += self.POLL_INTERVAL
            remaining_wallets = self.all_vks - set(self.init_balances.keys())

            if elapsed > self.FETCH_BALANCES_TIMEOUT:
                raise Exception("Exceeded timeout of {} trying to fetch initial wallet balances!\nWallets retreived: {}"
                                "\nWallets remaining: {}".format(self.FETCH_BALANCES_TIMEOUT, self.init_balances, remaining_wallets))

            balances = await God.get_balances(self.session, remaining_wallets)
            self.init_balances.update(balances)

        self.log.test("All initial wallet balances fetched!".format(self.init_balances))

    # TODO increase num_blocks to 4 once 1 works --davis
    async def send_test_currency_txs(self, num_blocks=4):
        assert len(self.init_balances) == len(self.wallets), "Init balances not equal to length of wallet"
        num = self.TX_PER_BLOCK * num_blocks
        self.log.test("Sending {} random test transactions...".format(num))

        txs = []
        for _ in range(num):
            sender, receiver = random.sample(self.wallets, 2)
            amount = random.randint(1, 10000)
            tx = God.create_currency_tx(sender, receiver, amount, self.STAMPS_AMOUNT)
            txs.append(tx)
            # self.deltas[sender[1]] -= (amount + self.STAMPS_AMOUNT)
            self.deltas[sender[1]] -= amount
            self.deltas[receiver[1]] += amount

        await God.async_send_txs(txs, self.session)
        self.log.test("Finish sending {} test transactions".format(num))

    async def assert_balances_updated(self):
        self.log.test("Starting assertion check for {} updated balances...".format(len(self.deltas)))

        elapsed = 0
        correct_wallets = set()
        latest_balances = {}  # Just for debugging info

        while len(correct_wallets) != len(self.deltas):
            remaining_vks = set(self.deltas.keys()) - correct_wallets

            if elapsed > self.ASSERT_BALANCES_TIMEOUT:
                raise Exception("Exceeded timeout of {} waiting for wallets to update!\nRemaining Wallets: {}\nLatest "
                                "Balances: {}\nDeltas: {}".format(self.ASSERT_BALANCES_TIMEOUT, remaining_vks,
                                                                  latest_balances, self.deltas))

            updated_balances = await God.get_balances(self.session, remaining_vks)
            for vk in updated_balances:
                expected_amount = max(0, self.init_balances[vk] + self.deltas[vk])
                actual_amount = updated_balances[vk]
                latest_balances[vk] = actual_amount
                if expected_amount == actual_amount:
                    correct_wallets.add(vk)
                else:
                    if elapsed >= self.ASSERT_BALANCES_TIMEOUT/2:
                        self.log.warning("Balance {} does not match expected balance {} for vk {}"
                                         .format(actual_amount, expected_amount, vk))

            await asyncio.sleep(self.POLL_INTERVAL)
            elapsed += self.POLL_INTERVAL

        self.log.success("Assertions for {} balance deltas passed!".format(len(self.deltas)))



"""
Utilities to build sets of signing keys and verifying keys for TestNet
"""
import cilantro_ee
import configparser
from cilantro_ee.core.crypto import wallet
from cilantro_ee.core.logger.base import get_logger
import json, os
from os.path import join

log = get_logger("TestnetNodeBuilder")

DEFAULT_TESTNET_FILE_NAME = '4-4-4.json'
TESTNET_JSON_DIR = os.path.dirname(cilantro_ee.__path__[-1]) + '/constitutions/test'
PUBLIC_JSON_DIR = os.path.dirname(cilantro_ee.__path__[-1]) + '/constitutions/public'
# print("TESTNET JSON DIR: {}".format(TESTNET_JSON_DIR))

TESTNET_KEY = 'testnet'
TESTNET_JSON_KEY = 'testnet_json_file_name'
CONFIG_FILE_NAME = 'testnet.ini'
CONFIG_FILE_PATH = join(TESTNET_JSON_DIR, CONFIG_FILE_NAME)


def set_testnet_config(testnet_json_file='4-4-4.json'):
    config = configparser.ConfigParser()
    config[TESTNET_KEY] = {TESTNET_JSON_KEY: testnet_json_file}

    with open(CONFIG_FILE_PATH, 'w+') as f:
        config.write(f)

    from cilantro_ee.constants.testnet import set_testnet_nodes
    set_testnet_nodes()


def get_testnet_config():
    with open(get_testnet_json_path(), 'r') as f:
        return json.load(f)


def get_config_filename():
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    return config[TESTNET_KEY][TESTNET_JSON_KEY]


def read_public_constitution(filename) -> dict:
    fpath = PUBLIC_JSON_DIR + '/' + filename
    assert os.path.exists(fpath), "No public constitution file found at path {}".format(fpath)
    with open(fpath) as f:
        return json.load(f)


def get_testnet_json_path():
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    return join(TESTNET_JSON_DIR, config[TESTNET_KEY][TESTNET_JSON_KEY])




from unittest import TestCase
from cilantro_ee.core.sockets import services
from cilantro_ee.core.crypto.wallet import Wallet

from cilantro_ee.services.block_fetch import BlockFetcher
from cilantro_ee.services.block_server import BlockServer
from cilantro_ee.core.messages.message import Message
from cilantro_ee.core.messages.message_type import MessageType
from cilantro_ee.core import canonical
import secrets
from cilantro_ee.services.storage.master import CilantroStorageDriver

from cilantro_ee.core.top import TopBlockManager
from cilantro_ee.core.sockets.socket_book import SocketBook

import time
import zmq.asyncio
import zmq
import asyncio
import hashlib
from tests import random_txs


class FakeTopBlockManager:
    def __init__(self, height, hash_):
        self.height = height
        self.hash_ = hash_

    def get_latest_block_hash(self):
        return self.hash_

    def get_latest_block_number(self):
        return self.height


class FakeSocketBook:
    def __init__(self, network=None, phonebook_function: callable = None):
        self.network = network
        self.phonebook_function = phonebook_function
        self.sockets = {}

    async def refresh(self):
        self.sockets = self.phonebook_function()


class FakeBlockDriver:
    def __init__(self, block_dict):
        self.block_dict = block_dict

    def get_block(self, i):
        return self.block_dict

class FakeBlockReciever:
    def __init__(self):
        self.blocks = {}

    def put(self, d):
        self.blocks[d['blockNum']] = d

async def stop_server(s, timeout):
    await asyncio.sleep(timeout)
    s.stop()


async def pause(f, time):
    await asyncio.sleep(time)
    return await f()


class TestBlockFetcher(TestCase):
    def setUp(self):
        self.ctx = zmq.asyncio.Context()
        self.t = TopBlockManager()

    def tearDown(self):
        self.ctx.destroy()
        self.t.driver.flush()

    def test_get_latest_block_height(self):
        w = Wallet()
        m = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10000'),
                        wallet=w,
                        ctx=self.ctx,
                        linger=500,
                        poll_timeout=100,
                        top=FakeTopBlockManager(101, 'abcd'))

        f = BlockFetcher(wallet=Wallet(), ctx=self.ctx)

        tasks = asyncio.gather(
            m.serve(),
            f.get_latest_block_height(services._socket('tcp://127.0.0.1:10000')),
            stop_server(m, 0.1),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertEqual(res[1], 101)

    def test_get_consensus_on_block_height(self):
        w1 = Wallet()
        m1 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10000'),
                         wallet=w1,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'))

        w2 = Wallet()
        m2 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10001'),
                         wallet=w2,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'))

        w3 = Wallet()
        m3 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10002'),
                         wallet=w3,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'))

        w4 = Wallet()
        m4 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10003'),
                         wallet=w4,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(90, 'abcd'))

        def get_sockets():
            return {
                'a': services._socket('tcp://127.0.0.1:10000'),
                'b': services._socket('tcp://127.0.0.1:10001'),
                'c': services._socket('tcp://127.0.0.1:10002'),
                'd': services._socket('tcp://127.0.0.1:10003'),
            }

        sock_book = FakeSocketBook(None, get_sockets)

        f = BlockFetcher(wallet=Wallet(), ctx=self.ctx, masternode_sockets=sock_book)

        tasks = asyncio.gather(
            m1.serve(),
            m2.serve(),
            m3.serve(),
            m4.serve(),
            f.find_missing_block_indexes(),
            stop_server(m1, 0.1),
            stop_server(m2, 0.1),
            stop_server(m3, 0.1),
            stop_server(m4, 0.1),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        self.assertEqual(res[4], 101)

    def store_blocks(self, c, i, initial_hash=b'\x00' * 32):
        current_hash = initial_hash
        for _i in range(i):
            block = random_txs.random_block(block_num=_i)
            d = canonical.block_from_subblocks([s for s in block.subBlocks], previous_hash=current_hash, block_num=_i)

            d['blockOwners'] = [secrets.token_bytes(32) for _ in range(12)]

            c.put(d)

            del d['_id']
            del d['blockOwners']

            current_hash = d['blockHash']

    def test_fetch_block_from_master(self):
        # Setup Mongo
        w = Wallet()
        c = CilantroStorageDriver(key=w.sk.encode())
        c.drop_collections()

        # Store 20 blocks
        self.store_blocks(c, 1)

        w1 = Wallet()
        m1 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10000'),
                         wallet=w1,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=c)

        def get_sockets():
            return {
                'a': services._socket('tcp://127.0.0.1:10000'),
            }

        sock_book = FakeSocketBook(None, get_sockets)

        f = BlockFetcher(wallet=Wallet(), ctx=self.ctx, masternode_sockets=sock_book)

        tasks = asyncio.gather(
            m1.serve(),
            f.get_block_from_master(0, services._socket('tcp://127.0.0.1:10000')),
            stop_server(m1, 0.3),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        block_dict = c.get_block(0)

        del block_dict['blockOwners']

        got_block = res[1]

        got = canonical.block_from_subblocks([s for s in got_block.subBlocks], previous_hash=b'\x00' * 32, block_num=0)

        self.assertDictEqual(block_dict, got)

    def test_fetch_block_from_multiple_masters_where_some_are_corrupted(self):
        w = Wallet()
        c = CilantroStorageDriver(key=w.sk.encode())
        c.drop_collections()

        # Store 20 blocks
        self.store_blocks(c, 1)

        # Good one
        w1 = Wallet()
        m1 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10000'),
                         wallet=w1,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=c)

        # Bad Ones
        bad_block = canonical.block_from_subblocks([s for s in random_txs.random_block().subBlocks],
                                                   previous_hash=b'\x01' * 32,
                                                   block_num=0)

        bad_block['blockOwners'] = [secrets.token_bytes(32) for _ in range(30)]

        d = FakeBlockDriver(bad_block)
        w2 = Wallet()
        m2 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10001'),
                         wallet=w2,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=d)

        w3 = Wallet()
        m3 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10002'),
                         wallet=w3,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=d)

        def get_sockets():
            return {
                'b': services._socket('tcp://127.0.0.1:10001'),
                'c': services._socket('tcp://127.0.0.1:10002'),
                'a': services._socket('tcp://127.0.0.1:10000'),
            }

        sock_book = FakeSocketBook(None, get_sockets)
        f = BlockFetcher(wallet=Wallet(), ctx=self.ctx, masternode_sockets=sock_book)

        tasks = asyncio.gather(
            m1.serve(),
            m2.serve(),
            m3.serve(),
            f.find_valid_block(0, latest_hash=b'\x00' * 32, timeout=3000),
            stop_server(m1, 0.7),
            stop_server(m2, 0.7),
            stop_server(m3, 0.7),
        )

        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(tasks)

        block_dict = c.get_block(0)

        del block_dict['blockOwners']

        got_block = res[3]

        got = canonical.block_from_subblocks([s for s in got_block.subBlocks], previous_hash=b'\x00' * 32, block_num=0)

        self.assertDictEqual(block_dict, got)

    def test_fetch_multiple_blocks_works_with_good_actors(self):
        w = Wallet()
        c = CilantroStorageDriver(key=w.sk.encode())
        c.drop_collections()

        # Store 20 blocks
        self.store_blocks(c, 10)

        # Good one
        w1 = Wallet()
        m1 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10000'),
                         wallet=w1,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=c)

        w2 = Wallet()
        m2 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10001'),
                         wallet=w2,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=c)

        w3 = Wallet()
        m3 = BlockServer(socket_id=services._socket('tcp://127.0.0.1:10002'),
                         wallet=w3,
                         ctx=self.ctx,
                         linger=500,
                         poll_timeout=100,
                         top=FakeTopBlockManager(101, 'abcd'),
                         driver=c)

        def get_sockets():
            return {
                'b': services._socket('tcp://127.0.0.1:10001'),
                'c': services._socket('tcp://127.0.0.1:10002'),
                'a': services._socket('tcp://127.0.0.1:10000'),
            }

        sock_book = FakeSocketBook(None, get_sockets)
        fake_driver = FakeBlockReciever()
        f = BlockFetcher(wallet=Wallet(), ctx=self.ctx, masternode_sockets=sock_book, blocks=fake_driver)

        tasks = asyncio.gather(
            m1.serve(),
            m2.serve(),
            m3.serve(),
            f.fetch_blocks(latest_block_available=9),
            stop_server(m1, 1),
            stop_server(m2, 1),
            stop_server(m3, 1),
        )

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)

        last_hash = b'\x00' * 32
        for i in range(10):
            block_dict = c.get_block(i)

            del block_dict['blockOwners']

            got = canonical.block_from_subblocks([s for s in fake_driver.blocks[i]['subBlocks']], previous_hash=last_hash,
                                                 block_num=i)

            last_hash = block_dict['blockHash']

            self.assertDictEqual(block_dict, got)
